# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import NotExistent
from aiida.orm import Group, Node
from aiida.orm.utils import load_group, load_node


class TestOrmUtils(AiidaTestCase):

    def test_load_group(self):
        """
        Test the functionality of load_group
        """
        name = 'groupie'
        group = Group(name=name).store()

        # Load through uuid
        loaded_group = load_group(group.uuid)
        self.assertEquals(loaded_group.uuid, group.uuid)

        # Load through pk
        loaded_group = load_group(group.pk)
        self.assertEquals(loaded_group.uuid, group.uuid)

        # Load through uuid explicitly
        loaded_group = load_group(uuid=group.uuid)
        self.assertEquals(loaded_group.uuid, group.uuid)

        # Load through pk explicitly
        loaded_group = load_group(pk=group.pk)
        self.assertEquals(loaded_group.uuid, group.uuid)

        # Load through partial uuid
        loaded_group = load_group(group.uuid[:2])
        self.assertEquals(loaded_group.uuid, group.uuid)

        # Load through partial uuid
        loaded_group = load_group(group.uuid[:10])
        self.assertEquals(loaded_group.uuid, group.uuid)

        with self.assertRaises(NotExistent):
            load_group('non-existent-uuid')


    def test_load_node(self):
        """
        Test the functionality of load_node
        """
        node = Node().store()

        # Load through uuid
        loaded_node = load_node(node.uuid)
        self.assertEquals(loaded_node.uuid, node.uuid)

        # Load through pk
        loaded_node = load_node(node.pk)
        self.assertEquals(loaded_node.uuid, node.uuid)

        # Load through uuid explicitly
        loaded_node = load_node(uuid=node.uuid)
        self.assertEquals(loaded_node.uuid, node.uuid)

        # Load through pk explicitly
        loaded_node = load_node(pk=node.pk)
        self.assertEquals(loaded_node.uuid, node.uuid)

        # Load through partial uuid
        loaded_node = load_node(node.uuid[:2])
        self.assertEquals(loaded_node.uuid, node.uuid)

        # Load through partial uuid
        loaded_node = load_node(node.uuid[:10])
        self.assertEquals(loaded_node.uuid, node.uuid)

        with self.assertRaises(NotExistent):
            load_group('non-existent-uuid')


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import json
from aiida.orm import Group, Node
from aiida.utils.serialize import serialize_data, deserialize_data
from aiida.backends.testbase import AiidaTestCase


class TestSerialize(AiidaTestCase):

    def test_serialize_round_trip(self):
        """
        Test the serialization of a dictionary with Nodes in various data structure
        Also make sure that the serialized data is json-serializable
        """
        node_a = Node().store()
        node_b = Node().store()

        data = {
            'test': 1,
            'list': [1, 2, 3, node_a],
            'dict': {
                ('Si',): node_b,
                'foo': 'bar'
            },
            'baz': 'aar'
        }

        serialized_data = serialize_data(data)
        json_dumped = json.dumps(serialized_data)
        deserialized_data = deserialize_data(serialized_data)

        # For now manual element-for-element comparison until we come up with general
        # purpose function that can equate two node instances properly
        self.assertEqual(data['test'], deserialized_data['test'])
        self.assertEqual(data['baz'], deserialized_data['baz'])
        self.assertEqual(data['list'][:3], deserialized_data['list'][:3])
        self.assertEqual(data['list'][3].uuid, deserialized_data['list'][3].uuid)
        self.assertEqual(data['dict'][('Si',)].uuid, deserialized_data['dict'][('Si',)].uuid)

    def test_serialize_group(self):
        """
        Test that serialization and deserialization of Groups works.
        Also make sure that the serialized data is json-serializable
        """
        group_name = 'groupie'
        group_a = Group(name=group_name).store()

        data = {
            'group': group_a
        }

        serialized_data = serialize_data(data)
        json_dumped = json.dumps(serialized_data)
        deserialized_data = deserialize_data(serialized_data)

        self.assertEqual(data['group'].uuid, deserialized_data['group'].uuid)
        self.assertEqual(data['group'].name, deserialized_data['group'].name)


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Defines the migrations on the config.json file."""

from aiida.common.exceptions import ConfigurationVersionError
from ._migrations import (
    _MIGRATION_LOOKUP, CURRENT_CONFIG_VERSION, OLDEST_COMPATIBLE_CONFIG_VERSION
)

VERSION_KEY = 'CONFIG_VERSION'
CURRENT_KEY = 'CURRENT'
OLDEST_KEY = 'OLDEST_COMPATIBLE'

__all__ = ['check_and_migrate_config', 'add_config_version']

def add_config_version(
        config,
        current_version=CURRENT_CONFIG_VERSION,
        oldest_version=OLDEST_COMPATIBLE_CONFIG_VERSION
    ):
    """Injects the current and oldest compatible version numbers into the config."""
    config[VERSION_KEY] = {CURRENT_KEY: current_version, OLDEST_KEY: oldest_version}

def check_and_migrate_config(config, store=True):
    """
    Checks if the config needs to be migrated, and performs the migration if needed.
    """
    if config_needs_migrating(config):
        config = migrate_config(config)
        from aiida.common.setup import store_config
        if store:
            store_config(config)
    return config

def config_needs_migrating(config):
    """Checks if the config needs to be migrated."""
    current, oldest = _get_config_version(config)
    if oldest > CURRENT_CONFIG_VERSION:
        raise ConfigurationVersionError("The configuration in 'config.json' is generated by a newer version of AiiDA, and not compatible with the current version.")
    return CURRENT_CONFIG_VERSION > current

def _get_config_version(config):
    """Gets the current and oldest compatible versions from the config."""
    try:
        versions = config[VERSION_KEY]
        current = versions[CURRENT_KEY]
        oldest = versions[OLDEST_KEY]
    except KeyError:
        current = 0
        oldest = 0
    return current, oldest

def migrate_config(config):
    """Runs the migration functions to update the config to the current version."""
    current, _ = _get_config_version(config)
    while current < CURRENT_CONFIG_VERSION:
        config = _MIGRATION_LOOKUP[current].apply(config)
        current, _ = _get_config_version(config)
    return config



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.backends.sqlalchemy.models.comment import DbComment
from aiida.orm.implementation.general.comment import AbstractComment
from aiida.common.exceptions import NotExistent, MultipleObjectsError




class Comment(AbstractComment):

    def __init__(self, **kwargs):

        # If no arguments are passed, then create a new DbComment
        if not kwargs:
            self.dbcomment = DbComment()

        # If a DbComment is passed as argument. Just use it and
        # wrap it with a Comment object
        elif 'dbcomment' in kwargs:
            # When a dbcomment is passed as argument, then no other arguments
            # should be passed.
            if len(kwargs) > 1:
                raise ValueError("When a DbComment is passed as argument, no"
                                 "further arguments are accepted.")
            dbcomment = kwargs.pop('dbcomment')
            if not isinstance(dbcomment, DbComment):
                raise ValueError("Expected a DbComment. Object of a different"
                                 "class was given as argument.")
            self.dbcomment = dbcomment

        else:
            id = kwargs.pop('id', None)
            if id is None:
                id = kwargs.pop('pk', None)
            user = kwargs.pop('user', None)
            dbnode = kwargs.pop('dbnode', None)

            # Constructing the default query
            dbcomment_query = DbComment.query

            # If an id is specified then we add it to the query
            if id is not None:
                dbcomment_query = dbcomment_query.filter_by(id=id)

            # If a user is specified then we add it to the query
            if user is not None:
                dbcomment_query = dbcomment_query.filter_by(user=user)

            # If a dbnode is specified then we add it to the query
            if dbnode is not None:
                dbcomment_query = dbcomment_query.filter_by(dbnode=dbnode)

            ccount = dbcomment_query.count()
            if ccount > 1:
                raise MultipleObjectsError(
                    "The arguments that you specified were too vague. More "
                    "than one comments with this data were found in the "
                    "database")
            elif ccount == 0:
                raise NotExistent("No comments were found with the given "
                                  "arguments")

            self.dbcomment = dbcomment_query.first()

    @property
    def pk(self):
        return self.dbcomment.id

    @property
    def id(self):
        return self.dbcomment.id

    @property
    def to_be_stored(self):
        return self.dbcomment.id is None

    @property
    def uuid(self):
        return unicode(self.dbcomment.uuid)

    def get_ctime(self):
        return self.dbcomment.ctime

    def set_ctime(self, val):
        self.dbcomment.ctime = val
        if not self.to_be_stored:
            self.dbcomment.save()

    def get_mtime(self):
        return self.dbcomment.mtime

    def set_mtime(self, val):
        self.dbcomment.mtime = val
        if not self.to_be_stored:
            self.dbcomment.save()

    def get_user(self):
        return self.dbcomment.user

    def set_user(self, val):
        self.dbcomment.user = val
        if not self.to_be_stored:
            self.dbcomment.save()

    def get_content(self):
        return self.dbcomment.content

    def set_content(self, val):
        self.dbcomment.content = val
        if not self.to_be_stored:
            self.dbcomment.save()

    def delete(self):
        self.dbcomment.delete()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from abc import abstractmethod, abstractproperty, ABCMeta
from collections import namedtuple
from aiida.utils import timezone

ASCENDING = 1
DESCENDING = -1

OrderSpecifier = namedtuple("OrderSpecifier", ['field', 'direction'])


class Log(object):
    """
    This class represents the collection of logs and can be used to create
    and retrieve logs.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def create_entry(self, time, loggername, levelname, objname,
                     objpk=None, message="", metadata=None):
        """
        Create a log entry.

        :param time: The time of creation for the entry
        :type time: :class:`!datetime.datetime`
        :param loggername: The name of the logger that generated the entry
        :type loggername: basestring
        :param levelname: The log level
        :type levelname: basestring
        :param objname: The object name (if any) that emitted the entry
        :type objname: basestring
        :param objpk: The object id that emitted the entry
        :type objpk: int
        :param message: The message to log
        :type message: basestring
        :param metadata: Any (optional) metadata, should be JSON serializable dictionary
        :type metadata: :class:`dict`
        :return: An object implementing the log entry interface
        :rtype: :class:`aiida.orm.log.LogEntry`
        """
        pass

    def create_entry_from_record(self, record):
        """
        Helper function to create a log entry from a record created as by the
        python logging lobrary

        :param record: The record created by the logging module
        :type record: :class:`logging.record`
        :return: An object implementing the log entry interface
        :rtype: :class:`aiida.orm.log.LogEntry`
        """
        from datetime import datetime

        objpk = record.__dict__.get('objpk', None)
        objname = record.__dict__.get('objname', None)

        # Do not store if objpk and objname are not set
        if objpk is None or objname is None:
            return None

        return self.create_entry(
            time=timezone.make_aware(datetime.fromtimestamp(record.created)),
            loggername=record.name,
            levelname=record.levelname,
            objname=objname,
            objpk=objpk,
            message=record.getMessage(),
            metadata=record.__dict__
        )

    @abstractmethod
    def find(self, filter_by=None, order_by=None, limit=None):
        """
        Find all entries in the Log collection that conforms to the filter and
        optionally sort and/or apply a limit.

        :param filter_by: A dictionary of key value pairs where the entries have
            to match all the criteria (i.e. an AND operation)
        :type filter_by: :class:`dict`
        :param order_by: A list of tuples of type :class:`OrderSpecifier`
        :type order_by: list
        :param limit: The number of entries to retrieve
        :return: An iterable of the matching entries
        """
        pass

    @abstractmethod
    def delete_many(self, filter):
        """
        Delete all the log entries matching the given filter
        """
        pass


class LogEntry(object):
    __metaclass__ = ABCMeta

    @abstractproperty
    def id(self):
        """
        Get the primary key of the entry

        :return: The entry primary key
        :rtype: int
        """
        pass

    @abstractproperty
    def time(self):
        """
        Get the time corresponding to the entry

        :return: The entry timestamp
        :rtype: :class:`!datetime.datetime`
        """
        pass

    @abstractproperty
    def loggername(self):
        """
        The name of the logger that created this entry

        :return: The entry loggername
        :rtype: basestring
        """
        pass

    @abstractproperty
    def levelname(self):
        """
        The name of the log level

        :return: The entry log level name
        :rtype: basestring
        """
        pass

    @abstractproperty
    def objpk(self):
        """
        Get the id of the object that created the log entry

        :return: The entry timestamp
        :rtype: int
        """
        pass

    @abstractproperty
    def objname(self):
        """
        Get the name of the object that created the log entry

        :return: The entry object name
        :rtype: basestring
        """
        pass

    @abstractproperty
    def message(self):
        """
        Get the message corresponding to the entry

        :return: The entry message
        :rtype: basestring
        """
        pass

    @abstractproperty
    def metadata(self):
        """
        Get the metadata corresponding to the entry

        :return: The entry metadata
        :rtype: :class:`!json.json`
        """
        pass

    @abstractmethod
    def save(self):
        """
        Persist the log entry to the database

        :return: reference of self
        :rtype: :class: LogEntry
        """
        pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for testing components from the workflow engine"""
import plumpy
from aiida.work.processes import Process


class DummyProcess(Process):
    """A Process that does nothing when it runs."""

    @classmethod
    def define(cls, spec):
        super(DummyProcess, cls).define(spec)
        spec.inputs.dynamic = True
        spec.outputs.dynamic = True

    def _run(self):
        pass


class AddProcess(Process):
    """A simple Process that adds two integers."""

    @classmethod
    def define(cls, spec):
        super(AddProcess, cls).define(spec)
        spec.input('a', required=True)
        spec.input('b', required=True)
        spec.output('result', required=True)

    def _run(self):
        self.out(self.inputs.a + self.inputs.b)


class BadOutput(Process):
    """A Process that emits an output that isn't part of the spec raising an exception."""

    @classmethod
    def define(cls, spec):
        super(BadOutput, cls).define(spec)
        spec.outputs.dynamic = True

    def _run(self):
        self.out("bad_output", 5)


class ExceptionProcess(Process):
    """A Process that raises a RuntimeError when run."""

    def _run(self):  # pylint: disable=no-self-use
        raise RuntimeError('CRASH')


class WaitProcess(Process):
    """A Process that waits until it is asked to continue."""

    def _run(self):
        return plumpy.Wait(self.next_step)

    def next_step(self):
        pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from abc import ABCMeta
from aiida.common.exceptions import InputValidationError, MultipleObjectsError, NotExistent
from aiida.plugins.factory import BaseFactory
from aiida.common.utils import abstractclassmethod

__all__ = ['CalculationFactory', 'DataFactory', 'WorkflowFactory', 'load_group', 
           'load_node', 'load_workflow', 'BackendDelegateWithDefault']


def CalculationFactory(entry_point):
    """
    Return the Calculation plugin class for a given entry point

    :param entry_point: the entry point name of the Calculation plugin
    """
    return BaseFactory('aiida.calculations', entry_point)


def DataFactory(entry_point):
    """
    Return the Data plugin class for a given entry point

    :param entry_point: the entry point name of the Data plugin
    """
    return BaseFactory('aiida.data', entry_point)


def WorkflowFactory(entry_point):
    """
    Return the Workflow plugin class for a given entry point

    :param entry_point: the entry point name of the Workflow plugin
    """
    return BaseFactory('aiida.workflows', entry_point)


def create_node_id_qb(node_id=None, pk=None, uuid=None, parent_class=None, query_with_dashes=True):
    """
    Returns the QueryBuilder instance set to retrieve AiiDA objects given their
    (parent)class and PK (in which case the object should be unique) or UUID
    or UUID starting pattern.

    :param node_id: PK (integer) or UUID (string) or a node
    :param pk: PK of a node
    :param uuid: UUID of a node, or the beginning of the uuid
    :param parent_class: if specified, looks only among objects that are instances of
    	a subclass of parent_class, otherwise among nodes
    :param bool query_with_dashes: Specific if uuid is passed, allows to
        put the uuid in the correct form. Default=True

    :return: a QueryBuilder instance
    """
    # This must be done inside here, because at import time the profile
    # must have been already loaded. If you put it at the module level,
    # the implementation is frozen to the default one at import time.
    from aiida.orm.implementation import Node
    from aiida.orm.querybuilder import QueryBuilder

    # First checking if the inputs are valid:
    inputs_provided = [val is not None for val in (node_id, pk, uuid)].count(True)
    if inputs_provided == 0:
        raise InputValidationError("one of the parameters 'node_id', 'pk' and 'uuid' has to be supplied")
    elif inputs_provided > 1:
        raise InputValidationError("only one of parameters 'node_id', 'pk' and 'uuid' has to be supplied")

    # In principle, I can use this function to fetch any kind of AiiDA object,
    # but if I don't specify anything, I assume that I am looking for nodes
    class_ = parent_class or Node

    # The logic is as follows: If pk is specified I will look for the pk
    # if UUID is specified for the uuid.
    # node_id can either be string -> uuid or an integer -> pk
    # Checking first if node_id specified
    if node_id is not None:
        if isinstance(node_id, (str, unicode)):
            uuid = node_id
        elif isinstance(node_id, int):
            pk = node_id
        else:
            raise TypeError("'node_id' has to be either string, unicode or "
                                 "integer, {} given".format(type(node_id)))

    # Check whether uuid, if supplied, is a string
    if uuid is not None:
        if not isinstance(uuid,(str, unicode)):
            raise TypeError("'uuid' has to be string or unicode")
    # Or whether the pk, if provided, is an integer
    elif pk is not None:
        if not isinstance(pk, int):
            raise TypeError("'pk' has to be an integer")
    else:
        # I really shouldn't get here
        assert True,  "Neither pk  nor uuid was provided"

    qb = QueryBuilder()
    qb.append(class_, tag='node')

    if pk:
        qb.add_filter('node',  {'id': pk})
    elif uuid:
        # Put back dashes in the right place
        start_uuid = uuid.replace('-', '')
        # TODO (only if it brings any speed advantage) add a check on the number of characters
        # to recognize if the uuid pattern is complete. If so, the filter operator can be '=='
        if query_with_dashes:
            # Essential that this is ordered from largest to smallest!
            for dash_pos in [20, 16, 12, 8]:
                if len(start_uuid) > dash_pos:
                    start_uuid = '{}-{}'.format(
                        start_uuid[:dash_pos], start_uuid[dash_pos:]
                        )

        qb.add_filter('node', {'uuid': {'like': '{}%'.format(start_uuid)}})

    return qb


def load_group(group_id=None, pk=None, uuid=None, query_with_dashes=True):
    """
    Load a group by its pk or uuid

    :param group_id: pk (integer) or uuid (string) of a group
    :param pk: pk of a group
    :param uuid: uuid of a group, or the beginning of the uuid
    :param bool query_with_dashes: allow to query for a uuid with dashes (default=True)
    :returns: the requested group if existing and unique
    :raise InputValidationError: if none or more than one of the arguments are supplied
    :raise TypeError: if the wrong types are provided
    :raise NotExistent: if no matching Node is found.
    :raise MultipleObjectsError: if more than one Node was found
    """
    from aiida.orm import Group

    kwargs = {
        'node_id': group_id,
        'pk': pk,
        'uuid': uuid,
        'parent_class': Group,
        'query_with_dashes': query_with_dashes
    }

    qb = create_node_id_qb(**kwargs)
    qb.add_projection('node', '*')
    qb.limit(2)

    try:
        return qb.one()[0]
    except MultipleObjectsError:
        raise MultipleObjectsError('More than one group found. Provide longer starting pattern for uuid.')
    except NotExistent:
        raise NotExistent('No group was found')


def load_node(node_id=None, pk=None, uuid=None, parent_class=None, query_with_dashes=True):
    """
    Load a node by its pk or uuid

    :param node_id: PK (integer) or UUID (string) or a node
    :param pk: PK of a node
    :param uuid: UUID of a node, or the beginning of the uuid
    :param parent_class: if specified, checks whether the node loaded is a
        subclass of parent_class
    :param bool query_with_dashes: allow to query for a uuid with dashes (default=True)
    :returns: the requested node if existing, unique, and (sub)instance of parent_class
    :raise InputValidationError: if none or more than one of the arguments are supplied
    :raise TypeError: if the wrong types are provided
    :raise NotExistent: if no matching Node is found.
    :raise MultipleObjectsError: if more than one Node was found

    """
    from aiida.orm.implementation import Node

    # I can use this functions to load only nodes, i.e. not users, groups etc ...
    # If nothing is specified I assume the big granpa: Node!
    class_ = parent_class or Node
    if not issubclass(class_,  Node):
        raise TypeError("{} is not a subclass of {}".format(class_, Node))

    kwargs = {
        'node_id': node_id,
        'pk': pk,
        'uuid': uuid,
        'parent_class': parent_class,
        'query_with_dashes': query_with_dashes
    }

    qb = create_node_id_qb(**kwargs)
    qb.add_projection('node', '*')
    qb.limit(2)

    try:
        return qb.one()[0]
    except MultipleObjectsError:
        raise MultipleObjectsError('More than one node found. Provide longer starting pattern for uuid.')
    except NotExistent:
        raise NotExistent('No node was found')


def load_workflow(wf_id=None, pk=None, uuid=None):
    """
    Return an AiiDA workflow given PK or UUID.

    :param wf_id: PK (integer) or UUID (string) or UUID instance or a workflow
    :param pk: PK of a workflow
    :param uuid: UUID of a workflow
    :return: an AiiDA workflow
    :raises: ValueError if none or more than one of parameters is supplied
        or type of wf_id is neither string nor integer
    """
    # This must be done inside here, because at import time the profile
    # must have been already loaded. If you put it at the module level,
    # the implementation is frozen to the default one at import time.
    from aiida.orm.implementation import Workflow
    from uuid import UUID as uuid_type

    if int(wf_id is None) + int(pk is None) + int(uuid is None) == 3:
        raise ValueError("one of the parameters 'wf_id', 'pk' and 'uuid' "
                         "has to be supplied")
    if int(wf_id is None) + int(pk is None) + int(uuid is None) < 2:
        raise ValueError("only one of parameters 'wf_id', 'pk' and 'uuid' "
                         "has to be supplied")

    if wf_id is not None:
        if wf_id and isinstance(wf_id, uuid_type):
            wf_id = str(wf_id)

        if isinstance(wf_id, basestring):
            return Workflow.get_subclass_from_uuid(wf_id)
        elif isinstance(wf_id, int):
            return Workflow.get_subclass_from_pk(wf_id)
        else:
            raise ValueError("'wf_id' has to be either string, unicode, "
                             "integer or UUID instance, {} given".format(type(wf_id)))
    if pk is not None:
        if isinstance(pk, int):
            return Workflow.get_subclass_from_pk(pk)
        else:
            raise ValueError("'pk' has to be an integer")
    else:
        if uuid and isinstance(uuid, uuid_type):
            uuid = str(uuid)
        if isinstance(uuid, str) or isinstance(uuid, unicode):
            return Workflow.get_subclass_from_uuid(uuid)
        else:
            raise ValueError("'uuid' has to be a string, unicode or a UUID instance")


class BackendDelegateWithDefault(object):
    """
    This class is a helper to implement the delegation pattern [1] by
    delegating functionality (i.e. calling through) to the backend class
    which will do the actual work.

    [1] https://en.wikipedia.org/wiki/Delegation_pattern
    """
    __metaclass__ = ABCMeta

    _DEFAULT = None

    @abstractclassmethod
    def create_default(cls):
        raise NotImplementedError("The subclass should implement this")

    @classmethod
    def get_default(cls):
        if cls._DEFAULT is None:
            cls._DEFAULT = cls.create_default()
        return cls._DEFAULT

    def __init__(self, backend):
        self._backend = backend



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.orm.calculation.job import JobCalculation
from aiida.orm.calculation.work import WorkCalculation
from aiida.orm.calculation import *
from aiida.orm.data import *
from aiida.orm.utils import *
from aiida.orm.code import Code
from aiida.orm.computer import Computer, delete_computer
from aiida.orm.querybuilder import QueryBuilder
from aiida.orm.workflow import Workflow
from .authinfo import *
from .user import *
from aiida.orm.group import Group

__all__ = (['JobCalculation', 'WorkCalculation', 'Code', 'Computer',
            'CalculationFactory', 'DataFactory', 'WorkflowFactory',
            'QueryBuilder', 'Workflow', 'Group', 'delete_computer'] +
           calculation.__all__ + utils.__all__ + user.__all__ + authinfo.__all__)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import datetime
from django.db import transaction, IntegrityError
from django.db.models import Q
from aiida.common.utils import str_timedelta
from aiida.common.datastructures import sort_states, calc_states
from aiida.common.exceptions import ModificationNotAllowed, DbContentError
from aiida.orm.group import Group
from aiida.orm.implementation.django.calculation import Calculation
from aiida.orm.implementation.general.calculation.job import AbstractJobCalculation
from aiida.plugins.loader import get_plugin_type_from_type_string
from aiida.utils import timezone


class JobCalculation(AbstractJobCalculation, Calculation):

    def _set_state(self, state):
        """
        Set the state of the calculation.

        Set it in the DbCalcState to have also the uniqueness check.
        Moreover (except for the IMPORTED state) also store in the 'state'
        attribute, useful to know it also after importing, and for faster
        querying.

        .. todo:: Add further checks to enforce that the states are set
           in order?

        :param state: a string with the state. This must be a valid string,
          from ``aiida.common.datastructures.calc_states``.
        :raise: ModificationNotAllowed if the given state was already set.
        """
        super(JobCalculation, self)._set_state(state)

        from aiida.common.datastructures import sort_states
        from aiida.backends.djsite.db.models import DbCalcState

        if not self.is_stored:
            raise ModificationNotAllowed("Cannot set the calculation state "
                                         "before storing")

        if state not in calc_states:
            raise ValueError(
                "'{}' is not a valid calculation status".format(state))

        old_state = self.get_state()
        if old_state:
            state_sequence = [state, old_state]

            # sort from new to old: if they are equal, then it is a valid
            # advance in state (otherwise, we are going backwards...)
            if sort_states(state_sequence) != state_sequence:
                raise ModificationNotAllowed("Cannot change the state from {} "
                                             "to {}".format(old_state, state))

        try:
            with transaction.atomic():
                new_state = DbCalcState(dbnode=self.dbnode, state=state).save()
        except IntegrityError:
            raise ModificationNotAllowed(
                "Calculation pk= {} already transited through "
                "the state {}".format(self.pk, state))

        # For non-imported states, also set in the attribute (so that, if we
        # export, we can still see the original state the calculation had.
        if state != calc_states.IMPORTED:
            self._set_attr('state', state)

    def get_state(self, from_attribute=False):
        """
        Get the state of the calculation.

        .. note:: this method returns the NOTFOUND state if no state
          is found in the DB.

        .. note:: the 'most recent' state is obtained using the logic in the
          ``aiida.common.datastructures.sort_states`` function.

        .. todo:: Understand if the state returned when no state entry is found
          in the DB is the best choice.

        :param from_attribute: if set to True, read it from the attributes
          (the attribute is also set with set_state, unless the state is set
          to IMPORTED; in this way we can also see the state before storing).

        :return: a string. If from_attribute is True and no attribute is found,
          return None. If from_attribute is False and no entry is found in the
          DB, return the "NOTFOUND" state.
        """
        from aiida.backends.djsite.db.models import DbCalcState
        if from_attribute:
            return self.get_attr('state', None)
        else:
            if not self.is_stored:
                return calc_states.NEW
            else:
                this_calc_states = DbCalcState.objects.filter(
                    dbnode=self).values_list('state', flat=True)
                if not this_calc_states:
                    return None
                else:
                    try:
                        most_recent_state = sort_states(this_calc_states)[0]
                    except ValueError as e:
                        raise DbContentError("Error in the content of the "
                                             "DbCalcState table ({})".format(
                            e.message))

                    return most_recent_state



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import numpy

_default_epsilon_length = 1e-5
_default_epsilon_angle = 1e-5


def change_reference(reciprocal_cell, kpoints, to_cartesian=True):
    """
    Change reference system, from cartesian to crystal coordinates (units of b1,b2,b3) or viceversa.

    :param reciprocal_cell: a 3x3 array representing the cell lattice vectors in reciprocal space
    :param kpoints: a list of (3) point coordinates
    :return kpoints: a list of (3) point coordinates in the new reference
    """
    if not isinstance(kpoints, numpy.ndarray):
        raise ValueError('kpoints must be a numpy.array')

    transposed_cell = numpy.transpose(numpy.array(reciprocal_cell))
    if to_cartesian:
        matrix = transposed_cell
    else:
        matrix = numpy.linalg.inv(transposed_cell)

    # note: kpoints is a list Nx3, matrix is 3x3.
    # hence, first transpose kpoints, then multiply, finally transpose it back
    return numpy.transpose(numpy.dot(matrix, numpy.transpose(kpoints)))

def analyze_cell(cell=None, pbc=None):
    """
    A function executed by the __init__ or by set_cell.
    If a cell is set, properties like a1, a2, a3, cosalpha, reciprocal_cell are
    set as well, although they are not stored in the DB.
    :note: units are Angstrom for the cell parameters, 1/Angstrom for the
    reciprocal cell parameters.
    """
    if pbc is None:
        pbc = [True, True, True]

    dimension = sum(pbc)

    if cell is None:
        return {
            'reciprocal_cell': None,
            'dimension': dimension,
            'pbc': pbc
        }

    the_cell = numpy.array(cell)
    reciprocal_cell = 2. * numpy.pi * numpy.linalg.inv(the_cell).transpose()
    a1 = numpy.array(the_cell[0, :])  # units = Angstrom
    a2 = numpy.array(the_cell[1, :])  # units = Angstrom
    a3 = numpy.array(the_cell[2, :])  # units = Angstrom
    a = numpy.linalg.norm(a1)  # units = Angstrom
    b = numpy.linalg.norm(a2)  # units = Angstrom
    c = numpy.linalg.norm(a3)  # units = Angstrom
    b1 = reciprocal_cell[0, :]  # units = 1/Angstrom
    b2 = reciprocal_cell[1, :]  # units = 1/Angstrom
    b3 = reciprocal_cell[2, :]  # units = 1/Angstrom
    cosalpha = numpy.dot(a2, a3) / b / c
    cosbeta = numpy.dot(a3, a1) / c / a
    cosgamma = numpy.dot(a1, a2) / a / b


    result = {
        'a1': a1,
        'a2': a2,
        'a3': a3,
        'a': a,
        'b': b,
        'c': c,
        'b1': b1,
        'b2': b2,
        'b3': b3,
        'cosalpha': cosalpha,
        'cosbeta': cosbeta,
        'cosgamma': cosgamma,
        'dimension': dimension,
        'reciprocal_cell': reciprocal_cell,
        'pbc': pbc,
    }

    return result


def get_explicit_kpoints_path(value=None, cell=None, pbc=None, kpoint_distance=None, cartesian=False,
                     epsilon_length=_default_epsilon_length,
                     epsilon_angle=_default_epsilon_angle):
    """
    Set a path of kpoints in the Brillouin zone.

    :param value: description of the path, in various possible formats.

        None: automatically sets all irreducible high symmetry paths.
        Requires that a cell was set

        or::

          [('G','M'), (...), ...]
          [('G','M',30), (...), ...]
          [('G',(0,0,0),'M',(1,1,1)), (...), ...]
          [('G',(0,0,0),'M',(1,1,1),30), (...), ...]

    :param cell: 3x3 array representing the structure cell lattice vectors
    :param pbc: 3-dimensional array of booleans signifying the periodic boundary
        conditions along each lattice vector
    :param float kpoint_distance: parameter controlling the distance between
        kpoints. Distance is given in crystal coordinates, i.e. the distance
        is computed in the space of b1,b2,b3. The distance set will be the
        closest possible to this value, compatible with the requirement of
        putting equispaced points between two special points (since extrema
        are included).
    :param bool cartesian: if set to true, reads the coordinates eventually
        passed in value as cartesian coordinates. Default: False.
    :param float epsilon_length: threshold on lengths comparison, used
        to get the bravais lattice info. It has to be used if the
        user wants to be sure the right symmetries are recognized.
    :param float epsilon_angle: threshold on angles comparison, used
        to get the bravais lattice info. It has to be used if the
        user wants to be sure the right symmetries are recognized.

    :returns: point_coordinates, path, bravais_info, explicit_kpoints, labels
    """
    bravais_info = find_bravais_info(
        cell=cell, pbc=pbc,
        epsilon_length=epsilon_length,
        epsilon_angle=epsilon_angle
    )

    analysis = analyze_cell(cell, pbc)
    dimension = analysis['dimension']
    reciprocal_cell = analysis['reciprocal_cell']
    pbc = list(analysis['pbc'])

    if dimension == 0:
        # case with zero dimension: only gamma-point is set
        return [[0., 0., 0.]], None, bravais_info

    def _is_path_1(path):
        try:
            are_two = all([len(i) == 2 for i in path])
            if not are_two:
                return False

            for i in path:
                are_str = all([isinstance(b, basestring) for b in i])
                if not are_str:
                    return False
        except IndexError:
            return False
        return True

    def _is_path_2(path):
        try:
            are_three = all([len(i) == 3 for i in path])
            if not are_three:
                return False

            are_good = all([all([isinstance(b[0], basestring),
                                 isinstance(b[1], basestring),
                                 isinstance(b[2], int)])
                            for b in path])
            if not are_good:
                return False

            # check that at least two points per segment (beginning and end)
            points_num = [int(i[2]) for i in path]
            if any([i < 2 for i in points_num]):
                raise ValueError("Must set at least two points per path "
                                 "segment")

        except IndexError:
            return False
        return True

    def _is_path_3(path):
        # [('G',(0,0,0),'M',(1,1,1)), (...), ...]
        try:
            _ = len(path)
            are_four = all([len(i) == 4 for i in path])
            if not are_four:
                return False
            have_labels = all(all([isinstance(i[0], basestring), isinstance(i[2], basestring)]) for i in path)
            if not have_labels:
                return False
            for i in path:
                coord1 = [float(j) for j in i[1]]
                coord2 = [float(j) for j in i[3]]
                if len(coord1) != 3 or len(coord2) != 3:
                    return False
        except (TypeError, IndexError):
            return False
        return True

    def _is_path_4(path):
        # [('G',(0,0,0),'M',(1,1,1),30), (...), ...]
        try:
            _ = len(path)
            are_five = all([len(i) == 5 for i in path])
            if not are_five:
                return False
            have_labels = all(all([isinstance(i[0], basestring), isinstance(i[2], basestring)]) for i in path)
            if not have_labels:
                return False
            have_points_num = all([isinstance(i[4], int) for i in path])
            if not have_points_num:
                return False
            # check that at least two points per segment (beginning and end)
            points_num = [int(i[4]) for i in path]
            if any([i < 2 for i in points_num]):
                raise ValueError("Must set at least two points per path "
                                 "segment")
            for i in path:
                coord1 = [float(j) for j in i[1]]
                coord2 = [float(j) for j in i[3]]
                if len(coord1) != 3 or len(coord2) != 3:
                    return False
        except (TypeError, IndexError):
            return False
        return True

    def _num_points_from_coordinates(path, point_coordinates, kpoint_distance=None):
        # NOTE: this way of creating intervals ensures equispaced objects
        #       in crystal coordinates of b1,b2,b3
        distances = [numpy.linalg.norm(numpy.array(point_coordinates[i[0]]) -
                                       numpy.array(point_coordinates[i[1]])
                                       ) for i in path]

        if kpoint_distance is None:
            # Use max_points_per_interval as the default guess for automatically
            # guessing the number of points
            max_point_per_interval = 10
            max_interval = max(distances)
            try:
                points_per_piece = [max(2, int(max_point_per_interval * i / max_interval)) for i in distances]
            except ValueError:
                raise ValueError("The beginning and end of each segment in the "
                                 "path should be different.")

        else:
            points_per_piece = [max(2, int(distance / kpoint_distance))
                                for distance in distances]
        return points_per_piece

    if cartesian:
        if cell is None:
            raise ValueError("To use cartesian coordinates, a cell must "
                             "be provided")

    if kpoint_distance is not None:
        if kpoint_distance <= 0.:
            raise ValueError("kpoints_distance must be a positive float")

    if value is None:
        if cell is None:
            raise ValueError("Cannot set a path not even knowing the "
                             "kpoints or at least the cell")
        point_coordinates, path, bravais_info = get_kpoints_path(
            cell=cell, pbc=pbc, cartesian=cartesian,
            epsilon_length=epsilon_length,
            epsilon_angle=epsilon_angle)
        num_points = _num_points_from_coordinates(path, point_coordinates, kpoint_distance)

    elif _is_path_1(value):
        # in the form [('X','M'),(...),...]
        if cell is None:
            raise ValueError("Cannot set a path not even knowing the "
                             "kpoints or at least the cell")

        path = value
        point_coordinates, _, bravais_info = get_kpoints_path(
            cell=cell, pbc=pbc, cartesian=cartesian,
            epsilon_length=epsilon_length,
            epsilon_angle=epsilon_angle)
        num_points = _num_points_from_coordinates(path, point_coordinates, kpoint_distance)

    elif _is_path_2(value):
        # [('G','M',30), (...), ...]
        if cell is None:
            raise ValueError("Cannot set a path not even knowing the "
                             "kpoints or at least the cell")

        path = [(i[0], i[1]) for i in value]
        point_coordinates, _, bravais_info = get_kpoints_path(
            cell=cell, pbc=pbc, cartesian=cartesian,
            epsilon_length=epsilon_length,
            epsilon_angle=epsilon_angle)
        num_points = [i[2] for i in value]

    elif _is_path_3(value):
        # [('G',(0,0,0),'M',(1,1,1)), (...), ...]
        path = [(i[0], i[2]) for i in value]

        point_coordinates = {}
        for piece in value:
            if piece[0] in point_coordinates:
                if point_coordinates[piece[0]] != piece[1]:
                    raise ValueError("Different points cannot have the same label")
            else:
                if cartesian:
                    point_coordinates[piece[0]] = change_reference(
                        reciprocal_cell,
                        numpy.array([piece[1]]),
                        to_cartesian=False)[0]
                else:
                    point_coordinates[piece[0]] = piece[1]
            if piece[2] in point_coordinates:
                if point_coordinates[piece[2]] != piece[3]:
                    raise ValueError("Different points cannot have the same label")
            else:
                if cartesian:
                    point_coordinates[piece[2]] = change_reference(
                        reciprocal_cell,
                        numpy.array([piece[3]]),
                        to_cartesian=False)[0]
                else:
                    point_coordinates[piece[2]] = piece[3]

        num_points = _num_points_from_coordinates(path, point_coordinates, kpoint_distance)

    elif _is_path_4(value):
        # [('G',(0,0,0),'M',(1,1,1),30), (...), ...]
        path = [(i[0], i[2]) for i in value]

        point_coordinates = {}
        for piece in value:
            if piece[0] in point_coordinates:
                if point_coordinates[piece[0]] != piece[1]:
                    raise ValueError("Different points cannot have the same label")
            else:
                if cartesian:
                    point_coordinates[piece[0]] = change_reference(
                        reciprocal_cell,
                        numpy.array([piece[1]]),
                        to_cartesian=False)[0]
                else:
                    point_coordinates[piece[0]] = piece[1]
            if piece[2] in point_coordinates:
                if point_coordinates[piece[2]] != piece[3]:
                    raise ValueError("Different points cannot have the same label")
            else:
                if cartesian:
                    point_coordinates[piece[2]] = change_reference(
                        reciprocal_cell,
                        numpy.array([piece[3]]),
                        to_cartesian=False)[0]
                else:
                    point_coordinates[piece[2]] = piece[3]

        num_points = [i[4] for i in value]

    else:
        raise ValueError("Input format not recognized")

    explicit_kpoints = [tuple(point_coordinates[path[0][0]])]
    labels = [(0, path[0][0])]

    for count_piece, i in enumerate(path):
        ini_label = i[0]
        end_label = i[1]
        ini_coord = point_coordinates[ini_label]
        end_coord = point_coordinates[end_label]

        path_piece = zip(numpy.linspace(ini_coord[0], end_coord[0],
                                        num_points[count_piece]),
                         numpy.linspace(ini_coord[1], end_coord[1],
                                        num_points[count_piece]),
                         numpy.linspace(ini_coord[2], end_coord[2],
                                        num_points[count_piece]),
                         )

        for count, j in enumerate(path_piece):
            if all(numpy.array(explicit_kpoints[-1]) == j):
                continue  # avoid duplcates
            else:
                explicit_kpoints.append(j)

            # add labels for the first and last point
            if count == 0:
                labels.append((len(explicit_kpoints) - 1, ini_label))
            if count == len(path_piece) - 1:
                labels.append((len(explicit_kpoints) - 1, end_label))

    # I still have some duplicates in the labels: eliminate them
    sorted(set(labels), key=lambda x: x[0])

    return point_coordinates, path, bravais_info, explicit_kpoints, labels


def find_bravais_info(cell, pbc, epsilon_length=_default_epsilon_length,
                       epsilon_angle=_default_epsilon_angle):
    """
    Finds the Bravais lattice of the cell passed in input to the Kpoint class
    :note: We assume that the cell given by the cell property is the
    primitive unit cell.

    .. note:: in 3D, this implementation expects
       that the structure is already standardized according to the Setyawan
       paper. If this is not the case, the kpoints and band structure returned will be incorrect. The only case
       that is dealt correctly by the library is the case when axes are swapped, where the library correctly
       takes this swapping/rotation into account to assign kpoint labels and coordinates.


    :param cell: 3x3 array representing the structure cell lattice vectors
    :param pbc: 3-dimensional array of booleans signifying the periodic boundary
        conditions along each lattice vector
        passed in value as cartesian coordinates. Default: False.
    :param float epsilon_length: threshold on lengths comparison, used
        to get the bravais lattice info. It has to be used if the
        user wants to be sure the right symmetries are recognized.
    :param float epsilon_angle: threshold on angles comparison, used
        to get the bravais lattice info. It has to be used if the
        user wants to be sure the right symmetries are recognized.
    :return: a dictionary, with keys short_name, extended_name, index
            (index of the Bravais lattice), and sometimes variation (name of
            the variation of the Bravais lattice) and extra (a dictionary
            with extra parameters used by the get_kpoints_path method)
    """
    if cell is None:
        return None

    analysis = analyze_cell(cell, pbc)
    a1 = analysis['a1']
    a2 = analysis['a2']
    a3 = analysis['a3']
    a = analysis['a']
    b = analysis['b']
    c = analysis['c']
    cosa = analysis['cosalpha']
    cosb = analysis['cosbeta']
    cosc = analysis['cosgamma']
    dimension = analysis['dimension']
    pbc = list(pbc)

    # values of cosines at various angles
    _90 = 0.
    _60 = 0.5
    _30 = numpy.sqrt(3.) / 2.
    _120 = -0.5

    # NOTE: in what follows, I'm assuming the textbook order of alfa, beta and gamma

    # TODO: Maybe additional checks to see if the "correct" primitive
    # cell is used ? (there are other equivalent primitive
    # unit cells to the one expected here, typically for body-, c-, and
    # face-centered lattices)

    def l_are_equals(a, b):
        # function to compare lengths
        return abs(a - b) <= epsilon_length

    def a_are_equals(a, b):
        # function to compare angles (actually, cosines)
        return abs(a - b) <= epsilon_angle

    if dimension == 3:
        # =========================================#
        # 3D case -> 14 possible Bravais lattices #
        # =========================================#

        comparison_length = [l_are_equals(a, b), l_are_equals(b, c),
                             l_are_equals(c, a)]
        comparison_angles = [a_are_equals(cosa, cosb), a_are_equals(cosb, cosc),
                             a_are_equals(cosc, cosa)]

        if comparison_length.count(True) == 3:

            # needed for the body centered orthorhombic:
            orci_a = numpy.linalg.norm(a2 + a3)
            orci_b = numpy.linalg.norm(a1 + a3)
            orci_c = numpy.linalg.norm(a1 + a2)
            orci_the_a, orci_the_b, orci_the_c = sorted([orci_a, orci_b, orci_c])
            bco1 = - (-orci_the_a ** 2 + orci_the_b ** 2 + orci_the_c ** 2) / (4. * a ** 2)
            bco2 = - (orci_the_a ** 2 - orci_the_b ** 2 + orci_the_c ** 2) / (4. * a ** 2)
            bco3 = - (orci_the_a ** 2 + orci_the_b ** 2 - orci_the_c ** 2) / (4. * a ** 2)

            # ======================#
            # simple cubic lattice #
            # ======================#
            if comparison_angles.count(True) == 3 and a_are_equals(cosa, _90):
                bravais_info = {"short_name": "cub",
                                "extended_name": "cubic",
                                "index": 1,
                                "permutation": [0, 1, 2]
                                }
            # =====================#
            # face centered cubic #
            # =====================#
            elif comparison_angles.count(True) == 3 and a_are_equals(cosa, _60):
                bravais_info = {"short_name": "fcc",
                                "extended_name": "face centered cubic",
                                "index": 2,
                                "permutation": [0, 1, 2]
                                }
            # =====================#
            # body centered cubic #
            # =====================#
            elif comparison_angles.count(True) == 3 and a_are_equals(cosa, -1. / 3.):
                bravais_info = {"short_name": "bcc",
                                "extended_name": "body centered cubic",
                                "index": 3,
                                "permutation": [0, 1, 2]
                                }
            # ==============#
            # rhombohedral #
            # ==============#
            elif comparison_angles.count(True) == 3:
                # logical order is important, this check must come after the cubic cases
                bravais_info = {"short_name": "rhl",
                                "extended_name": "rhombohedral",
                                "index": 11,
                                "permutation": [0, 1, 2]
                                }
                if cosa > 0.:
                    bravais_info['variation'] = 'rhl1'
                    eta = (1. + 4. * cosa) / (2. + 4. * cosa)
                    bravais_info['extra'] = {'eta': eta,
                                             'nu': 0.75 - eta / 2.,
                                             }
                else:
                    bravais_info['variation'] = 'rhl2'
                    eta = 1. / (2. * (1. - cosa) / (1. + cosa))
                    bravais_info['extra'] = {'eta': eta,
                                             'nu': 0.75 - eta / 2.,
                                             }

            # ==========================#
            # body centered tetragonal #
            # ==========================#
            elif comparison_angles.count(True) == 1:  # two angles are the same
                bravais_info = {"short_name": "bct",
                                "extended_name": "body centered tetragonal",
                                "index": 5,
                                }
                if comparison_angles.index(True) == 0:  # alfa=beta
                    ref_ang = cosa
                    bravais_info["permutation"] = [0, 1, 2]
                elif comparison_angles.index(True) == 1:  # beta=gamma
                    ref_ang = cosb
                    bravais_info["permutation"] = [2, 0, 1]
                else:  # comparison_angles.index(True)==2: # gamma = alfa
                    ref_ang = cosc
                    bravais_info["permutation"] = [1, 2, 0]

                if ref_ang >= 0.:
                    raise ValueError("Problems on the definition of "
                                     "body centered tetragonal lattices")
                the_c = numpy.sqrt(-4. * ref_ang * (a ** 2))
                the_a = numpy.sqrt(2. * a ** 2 - (the_c ** 2) / 2.)

                if the_c < the_a:
                    bravais_info['variation'] = 'bct1'
                    bravais_info['extra'] = {'eta': (1. + (the_c / the_a) ** 2) / 4.}
                else:
                    bravais_info['variation'] = 'bct2'
                    bravais_info['extra'] = {'eta': (1. + (the_a / the_c) ** 2) / 4.,
                                             'csi': ((the_a / the_c) ** 2) / 2.,
                                             }

            # ============================#
            # body centered orthorhombic #
            # ============================#
            elif (any([a_are_equals(cosa, bco1), a_are_equals(cosb, bco1), a_are_equals(cosc, bco1)]) and
                      any([a_are_equals(cosa, bco2), a_are_equals(cosb, bco2), a_are_equals(cosc, bco2)]) and
                      any([a_are_equals(cosa, bco3), a_are_equals(cosb, bco3), a_are_equals(cosc, bco3)])
                  ):
                bravais_info = {"short_name": "orci",
                                "extended_name": "body centered orthorhombic",
                                "index": 8,
                                }
                if a_are_equals(cosa, bco1) and a_are_equals(cosc, bco3):
                    bravais_info['permutation'] = [0, 1, 2]
                if a_are_equals(cosa, bco1) and a_are_equals(cosc, bco2):
                    bravais_info['permutation'] = [0, 2, 1]
                if a_are_equals(cosa, bco3) and a_are_equals(cosc, bco2):
                    bravais_info['permutation'] = [1, 2, 0]
                if a_are_equals(cosa, bco2) and a_are_equals(cosc, bco3):
                    bravais_info['permutation'] = [1, 0, 2]
                if a_are_equals(cosa, bco2) and a_are_equals(cosc, bco1):
                    bravais_info['permutation'] = [2, 0, 1]
                if a_are_equals(cosa, bco3) and a_are_equals(cosc, bco1):
                    bravais_info['permutation'] = [2, 1, 0]

                bravais_info['extra'] = {'csi': (1. + (orci_the_a / orci_the_c) ** 2) / 4.,
                                         'eta': (1. + (orci_the_b / orci_the_c) ** 2) / 4.,
                                         'dlt': (orci_the_b ** 2 - orci_the_a ** 2) / (4. * orci_the_c ** 2),
                                         'mu': (orci_the_a ** 2 + orci_the_b ** 2) / (4. * orci_the_c ** 2),
                                         }

            # if it doesn't fall in the above, is triclinic
            else:
                bravais_info = {"short_name": "tri",
                                "extended_name": "triclinic",
                                "index": 14,
                                }
                # the check for triclinic variations is at the end of the method



        elif comparison_length.count(True) == 1:
            # ============#
            # tetragonal #
            # ============#
            if comparison_angles.count(True) == 3 and a_are_equals(cosa, _90):
                bravais_info = {"short_name": "tet",
                                "extended_name": "tetragonal",
                                "index": 4,
                                }
                if comparison_length[0] == True:
                    bravais_info["permutation"] = [0, 1, 2]
                if comparison_length[1] == True:
                    bravais_info["permutation"] = [2, 0, 1]
                if comparison_length[2] == True:
                    bravais_info["permutation"] = [1, 2, 0]
            # ====================================#
            # c-centered orthorombic + hexagonal #
            # ====================================#
            # alpha/=beta=gamma=pi/2
            elif (comparison_angles.count(True) == 1 and
                      any([a_are_equals(cosa, _90), a_are_equals(cosb, _90), a_are_equals(cosc, _90)])
                  ):
                if any([a_are_equals(cosa, _120), a_are_equals(cosb, _120), a_are_equals(cosc, _120)]):
                    bravais_info = {"short_name": "hex",
                                    "extended_name": "hexagonal",
                                    "index": 10,
                                    }
                else:
                    bravais_info = {"short_name": "orcc",
                                    "extended_name": "c-centered orthorhombic",
                                    "index": 9,
                                    }
                    if comparison_length[0] == True:
                        the_a1 = a1
                        the_a2 = a2
                    elif comparison_length[1] == True:
                        the_a1 = a2
                        the_a2 = a3
                    else:  # comparison_length[2]==True:
                        the_a1 = a3
                        the_a2 = a1
                    the_a = numpy.linalg.norm(the_a1 + the_a2)
                    the_b = numpy.linalg.norm(the_a1 - the_a2)
                    bravais_info['extra'] = {'csi': (1. + (the_a / the_b) ** 2) / 4.,
                                             }

                # TODO : re-check this case, permutations look weird
                if comparison_length[0] == True:
                    bravais_info["permutation"] = [0, 1, 2]
                if comparison_length[1] == True:
                    bravais_info["permutation"] = [2, 0, 1]
                if comparison_length[2] == True:
                    bravais_info["permutation"] = [1, 2, 0]
            # =======================#
            # c-centered monoclinic #
            # =======================#
            elif comparison_angles.count(True) == 1:
                bravais_info = {"short_name": "mclc",
                                "extended_name": "c-centered monoclinic",
                                "index": 13,
                                }
                # TODO : re-check this case, permutations look weird
                if comparison_length[0] == True:
                    bravais_info["permutation"] = [0, 1, 2]
                    the_ka = cosa
                    the_a1 = a1
                    the_a2 = a2
                    the_c = c
                if comparison_length[1] == True:
                    bravais_info["permutation"] = [2, 0, 1]
                    the_ka = cosb
                    the_a1 = a2
                    the_a2 = a3
                    the_c = a
                if comparison_length[2] == True:
                    bravais_info["permutation"] = [1, 2, 0]
                    the_ka = cosc
                    the_a1 = a3
                    the_a2 = a1
                    the_c = b

                the_b = numpy.linalg.norm(the_a1 + the_a2)
                the_a = numpy.linalg.norm(the_a1 - the_a2)
                the_cosa = 2. * numpy.linalg.norm(the_a1) / the_b * the_ka

                if a_are_equals(the_ka, _90):  # order matters: has to be before the check on mclc1
                    bravais_info['variation'] = 'mclc2'
                    csi = (2. - the_b * the_cosa / the_c) / (4. * (1. - the_cosa ** 2))
                    psi = 0.75 - the_a ** 2 / (4. * the_b * (1. - the_cosa ** 2))
                    bravais_info['extra'] = {'csi': csi,
                                             'eta': 0.5 + 2. * csi * the_c * the_cosa / the_b,
                                             'psi': psi,
                                             'phi': psi + (0.75 - psi) * the_b * the_cosa / the_c,
                                             }
                elif the_ka < 0.:
                    bravais_info['variation'] = 'mclc1'
                    csi = (2. - the_b * the_cosa / the_c) / (4. * (1. - the_cosa ** 2))
                    psi = 0.75 - the_a ** 2 / (4. * the_b * (1. - the_cosa ** 2))
                    bravais_info['extra'] = {'csi': csi,
                                             'eta': 0.5 + 2. * csi * the_c * the_cosa / the_b,
                                             'psi': psi,
                                             'phi': psi + (0.75 - psi) * the_b * the_cosa / the_c,
                                             }
                else:  # if the_ka>0.:
                    x = the_b * the_cosa / the_c + the_b ** 2 * (1. - the_cosa ** 2) / the_a ** 2
                    if a_are_equals(x, 1.):
                        bravais_info['variation'] = 'mclc4'  # order matters here too
                        mu = (1. + (the_b / the_a) ** 2) / 4.
                        dlt = the_b * the_c * the_cosa / (2. * the_a ** 2)
                        csi = mu - 0.25 + (1. - the_b * the_cosa / the_c) / (4. * (1. - the_cosa ** 2))
                        eta = 0.5 + 2. * csi * the_c * the_cosa / the_b
                        phi = 1. + eta - 2. * mu
                        psi = eta - 2. * dlt
                        bravais_info['extra'] = {'mu': mu,
                                                 'dlt': dlt,
                                                 'csi': csi,
                                                 'eta': eta,
                                                 'phi': phi,
                                                 'psi': psi,
                                                 }
                    elif x < 1.:
                        bravais_info['variation'] = 'mclc3'
                        mu = (1. + (the_b / the_a) ** 2) / 4.
                        dlt = the_b * the_c * the_cosa / (2. * the_a ** 2)
                        csi = mu - 0.25 + (1. - the_b * the_cosa / the_c) / (4. * (1. - the_cosa ** 2))
                        eta = 0.5 + 2. * csi * the_c * the_cosa / the_b
                        phi = 1. + eta - 2. * mu
                        psi = eta - 2. * dlt
                        bravais_info['extra'] = {'mu': mu,
                                                 'dlt': dlt,
                                                 'csi': csi,
                                                 'eta': eta,
                                                 'phi': phi,
                                                 'psi': psi,
                                                 }
                    elif x > 1.:
                        bravais_info['variation'] = 'mclc5'
                        csi = ((the_b / the_a) ** 2 + (1. - the_b * the_cosa / the_c) / (1. - the_cosa ** 2)) / 4.
                        eta = 0.5 + 2. * csi * the_c * the_cosa / the_b
                        mu = eta / 2. + the_b ** 2 / 4. / the_a ** 2 - the_b * the_c * the_cosa / 2. / the_a ** 2
                        nu = 2. * mu - csi
                        omg = (4. * nu - 1. - the_b ** 2 * (1. - the_cosa ** 2) / the_a ** 2) * the_c / (
                        2. * the_b * the_cosa)
                        dlt = csi * the_c * the_cosa / the_b + omg / 2. - 0.25
                        rho = 1. - csi * the_a ** 2 / the_b ** 2
                        bravais_info['extra'] = {'mu': mu,
                                                 'dlt': dlt,
                                                 'csi': csi,
                                                 'eta': eta,
                                                 'rho': rho,
                                                 }

            # if it doesn't fall in the above, is triclinic
            else:
                bravais_info = {"short_name": "tri",
                                "extended_name": "triclinic",
                                "index": 14,
                                }
                # the check for triclinic variations is at the end of the method



        else:  # if comparison_length.count(True)==0:

            fco1 = c ** 2 / numpy.sqrt((a ** 2 + c ** 2) * (b ** 2 + c ** 2))
            fco2 = a ** 2 / numpy.sqrt((a ** 2 + b ** 2) * (a ** 2 + c ** 2))
            fco3 = b ** 2 / numpy.sqrt((a ** 2 + b ** 2) * (b ** 2 + c ** 2))
            # ==============#
            # orthorhombic #
            # ==============#
            if comparison_angles.count(True) == 3:
                bravais_info = {"short_name": "orc",
                                "extended_name": "orthorhombic",
                                "index": 6,
                                }
                lens = [a, b, c]
                ind_a = lens.index(min(lens))
                ind_c = lens.index(max(lens))
                if ind_a == 0 and ind_c == 1:
                    bravais_info["permutation"] = [0, 2, 1]
                if ind_a == 0 and ind_c == 2:
                    bravais_info["permutation"] = [0, 1, 2]
                if ind_a == 1 and ind_c == 0:
                    bravais_info["permutation"] = [1, 2, 0]
                if ind_a == 1 and ind_c == 2:
                    bravais_info["permutation"] = [1, 0, 2]
                if ind_a == 2 and ind_c == 0:
                    bravais_info["permutation"] = [2, 1, 0]
                if ind_a == 2 and ind_c == 1:
                    bravais_info["permutation"] = [2, 0, 1]
            # ============#
            # monoclinic #
            # ============#
            elif (comparison_angles.count(True) == 1 and
                      any([a_are_equals(cosa, _90), a_are_equals(cosb, _90), a_are_equals(cosc, _90)])):
                bravais_info = {"short_name": "mcl",
                                "extended_name": "monoclinic",
                                "index": 12,
                                }
                lens = [a, b, c]
                # find the angle different from 90
                # then order (if possible) a<b<c
                if not a_are_equals(cosa, _90):
                    the_cosa = cosa
                    the_a = min(a, b)
                    the_b = max(a, b)
                    the_c = c
                    if lens.index(the_a) == 0:
                        bravais_info['permutation'] = [0, 1, 2]
                    else:
                        bravais_info['permutation'] = [1, 0, 2]
                elif not a_are_equals(cosb, _90):
                    the_cosa = cosb
                    the_a = min(a, c)
                    the_b = max(a, c)
                    the_c = b
                    if lens.index(the_a) == 0:
                        bravais_info['permutation'] = [0, 2, 1]
                    else:
                        bravais_info['permutation'] = [1, 2, 0]
                else:  # if not _are_equals(cosc,_90):
                    the_cosa = cosc
                    the_a = min(b, c)
                    the_b = max(b, c)
                    the_c = a
                    if lens.index(the_a) == 1:
                        bravais_info['permutation'] = [2, 0, 1]
                    else:
                        bravais_info['permutation'] = [2, 1, 0]
                eta = (1. - the_b * the_cosa / the_c) / (2. * (1. - the_cosa ** 2))
                bravais_info['extra'] = {'eta': eta,
                                         'nu': 0.5 - eta * the_c * the_cosa / the_b,
                                         }
            # ============================#
            # face centered orthorhombic #
            # ============================#
            elif (any([a_are_equals(cosa, fco1), a_are_equals(cosb, fco1), a_are_equals(cosc, fco1)]) and
                      any([a_are_equals(cosa, fco2), a_are_equals(cosb, fco2), a_are_equals(cosc, fco2)]) and
                      any([a_are_equals(cosa, fco3), a_are_equals(cosb, fco3), a_are_equals(cosc, fco3)])
                  ):
                bravais_info = {"short_name": "orcf",
                                "extended_name": "face centered orthorhombic",
                                "index": 7,
                                }

                lens = [a, b, c]
                ind_a1 = lens.index(max(lens))
                ind_a3 = lens.index(min(lens))
                if ind_a1 == 0 and ind_a3 == 2:
                    bravais_info['permutation'] = [0, 1, 2]
                    the_a1 = a1
                    the_a2 = a2
                    the_a3 = a3
                elif ind_a1 == 0 and ind_a3 == 1:
                    bravais_info['permutation'] = [0, 2, 1]
                    the_a1 = a1
                    the_a2 = a3
                    the_a3 = a2
                elif ind_a1 == 1 and ind_a3 == 2:
                    bravais_info['permutation'] = [1, 0, 2]
                    the_a1 = a2
                    the_a2 = a1
                    the_a3 = a3
                elif ind_a1 == 1 and ind_a3 == 0:
                    bravais_info['permutation'] = [2, 0, 1]
                    the_a1 = a3
                    the_a2 = a1
                    the_a3 = a2
                elif ind_a1 == 2 and ind_a3 == 1:
                    bravais_info['permutation'] = [1, 2, 0]
                    the_a1 = a2
                    the_a2 = a3
                    the_a3 = a1
                else:  # ind_a1 == 2 and ind_a3 == 0:
                    bravais_info['permutation'] = [2, 1, 0]
                    the_a1 = a3
                    the_a2 = a2
                    the_a3 = a1

                the_a = numpy.linalg.norm(- the_a1 + the_a2 + the_a3)
                the_b = numpy.linalg.norm(+ the_a1 - the_a2 + the_a3)
                the_c = numpy.linalg.norm(+ the_a1 + the_a2 - the_a3)

                fco4 = 1. / the_a ** 2 - 1. / the_b ** 2 - 1. / the_c ** 2
                # orcf3
                if a_are_equals(fco4, 0.):
                    bravais_info['variation'] = 'orcf3'  # order matters
                    bravais_info['extra'] = {'csi': (1. + (the_a / the_b) ** 2 - (the_a / the_c) ** 2) / 4.,
                                             'eta': (1. + (the_a / the_b) ** 2 + (the_a / the_c) ** 2) / 4.,
                                             }
                # orcf1
                elif fco4 > 0.:
                    bravais_info['variation'] = 'orcf1'
                    bravais_info['extra'] = {'csi': (1. + (the_a / the_b) ** 2 - (the_a / the_c) ** 2) / 4.,
                                             'eta': (1. + (the_a / the_b) ** 2 + (the_a / the_c) ** 2) / 4.,
                                             }
                # orcf2
                else:
                    bravais_info['variation'] = 'orcf2'
                    bravais_info['extra'] = {'eta': (1. + (the_a / the_b) ** 2 - (the_a / the_c) ** 2) / 4.,
                                             'dlt': (1. + (the_b / the_a) ** 2 + (the_b / the_c) ** 2) / 4.,
                                             'phi': (1. + (the_c / the_b) ** 2 - (the_c / the_a) ** 2) / 4.,
                                             }

            else:
                bravais_info = {"short_name": "tri",
                                "extended_name": "triclinic",
                                "index": 14,
                                }
        # ===========#
        # triclinic #
        # ===========#
        # still miss the variations of triclinic
        if bravais_info['short_name'] == 'tri':
            lens = [a, b, c]
            ind_a = lens.index(min(lens))
            ind_c = lens.index(max(lens))
            if ind_a == 0 and ind_c == 1:
                the_a = a
                the_b = c
                the_c = b
                the_cosa = cosa
                the_cosb = cosc
                the_cosc = cosb
                bravais_info["permutation"] = [0, 2, 1]
            if ind_a == 0 and ind_c == 2:
                the_a = a
                the_b = b
                the_c = c
                the_cosa = cosa
                the_cosb = cosb
                the_cosc = cosc
                bravais_info["permutation"] = [0, 1, 2]
            if ind_a == 1 and ind_c == 0:
                the_a = b
                the_b = c
                the_c = a
                the_cosa = cosb
                the_cosb = cosc
                the_cosc = cosa
                bravais_info["permutation"] = [1, 0, 2]
            if ind_a == 1 and ind_c == 2:
                the_a = b
                the_b = a
                the_c = c
                the_cosa = cosb
                the_cosb = cosa
                the_cosc = cosc
                bravais_info["permutation"] = [1, 0, 2]
            if ind_a == 2 and ind_c == 0:
                the_a = c
                the_b = b
                the_c = a
                the_cosa = cosc
                the_cosb = cosb
                the_cosc = cosa
                bravais_info["permutation"] = [2, 1, 0]
            if ind_a == 2 and ind_c == 1:
                the_a = c
                the_b = a
                the_c = b
                the_cosa = cosc
                the_cosb = cosa
                the_cosc = cosb
                bravais_info["permutation"] = [2, 0, 1]

            if the_cosa < 0. and the_cosb < 0.:
                if a_are_equals(the_cosc, 0.):
                    bravais_info['variation'] = 'tri2a'
                elif the_cosc < 0.:
                    bravais_info['variation'] = 'tri1a'
                else:
                    raise ValueError("Structure erroneously fell into the triclinic (a) case")
            elif the_cosa > 0. and the_cosb > 0.:
                if a_are_equals(the_cosc, 0.):
                    bravais_info['variation'] = 'tri2b'
                elif the_cosc > 0.:
                    bravais_info['variation'] = 'tri1b'
                else:
                    raise ValueError("Structure erroneously fell into the triclinic (b) case")
            else:
                raise ValueError("Structure erroneously fell into the triclinic case")

    elif dimension == 2:
        # ========================================#
        # 2D case -> 5 possible Bravais lattices #
        # ========================================#

        # find the two in-plane lattice vectors
        out_of_plane_index = pbc.index(False)  # the non-periodic dimension
        in_plane_indexes = list(set(range(3)) - set([out_of_plane_index]))
        # in_plane_indexes are the indexes of the two dimensions (e.g. [0,1])

        # build a length-2 list with the 2D cell lattice vectors
        list_vectors = ['a1', 'a2', 'a3']
        vectors = [eval(list_vectors[i]) for i in in_plane_indexes]
        # build a length-2 list with the norms of the 2D cell lattice vectors
        lens = [numpy.linalg.norm(v) for v in vectors]
        # cosine of the angle between the two primitive vectors
        list_angles = ['cosa', 'cosb', 'cosc']
        cosphi = eval(list_angles[out_of_plane_index])

        comparison_length = l_are_equals(lens[0], lens[1])
        comparison_angle_90 = a_are_equals(cosphi, _90)

        # ================#
        # square lattice #
        # ================#
        if comparison_angle_90 and comparison_length:
            bravais_info = {"short_name": "sq",
                            "extended_name": "square",
                            "index": 1,
                            }

        # =========================#
        # (primitive) rectangular #
        # =========================#
        elif comparison_angle_90:
            bravais_info = {"short_name": "rec",
                            "extended_name": "rectangular",
                            "index": 2,
                            }
            # set the order such that first_vector < second_vector in norm
            if lens[0] > lens[1]:
                in_plane_indexes.reverse()

        # ===========#
        # hexagonal #
        # ===========#
        # this has to be put before the centered-rectangular case
        elif (l_are_equals(lens[0], lens[1]) and a_are_equals(cosphi, _120)):
            bravais_info = {"short_name": "hex",
                            "extended_name": "hexagonal",
                            "index": 4,
                            }

        # ======================#
        # centered rectangular #
        # ======================#
        elif (comparison_length and
                  l_are_equals(numpy.dot(vectors[0] + vectors[1],
                                         vectors[0] - vectors[1]), 0.)):
            bravais_info = {"short_name": "recc",
                            "extended_name": "centered rectangular",
                            "index": 3,
                            }

        # =========#
        # oblique #
        # =========#
        else:
            bravais_info = {"short_name": "obl",
                            "extended_name": "oblique",
                            "index": 5,
                            }
            # set the order such that first_vector < second_vector in norm
            if lens[0] > lens[1]:
                in_plane_indexes.reverse()

        # the permutation is set such that p[2]=out_of_plane_index (third
        # new axis is always the non-periodic out-of-plane axis)
        # TODO: check that this (and the special points permutation of
        # coordinates) works also when the out-of-plane axis is not aligned
        # with one of the cartesian axis (I suspect that it doesn't...)
        permutation = in_plane_indexes + [out_of_plane_index]
        bravais_info["permutation"] = permutation

    elif dimension <= 1:
        # ====================================================#
        # 0D & 1D cases -> only one possible Bravais lattice #
        # ====================================================#
        if dimension == 1:
            # TODO: check that this (and the special points permutation of
            # coordinates) works also when the 1D axis is not aligned
            # with one of the cartesian axis (I suspect that it doesn't...)
            in_line_index = pbc.index(True)  # the only periodic dimension
            # the permutation is set such that p[0]=in_line_index (the 2 last
            # axes are always the non-periodic ones)
            permutation = [in_line_index] + list(set(range(3)) - set([in_line_index]))
        else:
            permutation = [0, 1, 2]

        bravais_info = {
            "short_name": "{}D".format(dimension),
            "extended_name": "{}D".format(dimension),
            "index": 1,
            "permutation": permutation,
        }

    return bravais_info



def get_kpoints_path(cell, pbc=None, cartesian=False,
                       epsilon_length=_default_epsilon_length,
                       epsilon_angle=_default_epsilon_angle):
    """
    Get the special point and path of a given structure.

    .. note:: in 3D, this implementation expects
       that the structure is already standardized according to the Setyawan
       paper. If this is not the case, the kpoints and band structure returned will be incorrect. The only case
       that is dealt correctly by the library is the case when axes are swapped, where the library correctly
       takes this swapping/rotation into account to assign kpoint labels and coordinates.

    - In 2D, coordinates are based on the paper:
      R. Ramirez and M. C. Bohm,  Int. J. Quant. Chem., XXX, pp. 391-411 (1986)

    - In 3D, coordinates are based on the paper:
      W. Setyawan, S. Curtarolo, Comp. Mat. Sci. 49, 299 (2010)

    :param cell: 3x3 array representing the structure cell lattice vectors
    :param pbc: 3-dimensional array of booleans signifying the periodic boundary
        conditions along each lattice vector
    :param cartesian: If true, returns points in cartesian coordinates.
        Crystal coordinates otherwise. Default=False
    :param epsilon_length: threshold on lengths comparison, used
        to get the bravais lattice info
    :param epsilon_angle: threshold on angles comparison, used
        to get the bravais lattice info
    :return special_points,path: special_points: a dictionary of
        point_name:point_coords key,values.

        path: the suggested path which goes through all high symmetry
        lines. A list of lists for all path segments.
        e.g. ``[('G','X'),('X','M'),...]``
        It's not necessarily a continuous line.
    :note: We assume that the cell given by the cell property is the
        primitive unit cell
    """
    # recognize which bravais lattice we are dealing with
    bravais_info = find_bravais_info(
        cell=cell, pbc=pbc,
        epsilon_length=epsilon_length,
        epsilon_angle=epsilon_angle
    )

    analysis = analyze_cell(cell, pbc)
    dimension = analysis['dimension']
    reciprocal_cell = analysis['reciprocal_cell']

    # pick the information about the special k-points.
    # it depends on the dimensionality and the Bravais lattice number.
    if dimension == 3:
        # 3D case: 14 Bravais lattices
        # simple cubic
        if bravais_info['index'] == 1:
            special_points = {'G': [0., 0., 0.],
                              'M': [0.5, 0.5, 0.],
                              'R': [0.5, 0.5, 0.5],
                              'X': [0., 0.5, 0.],
                              }
            path = [('G', 'X'),
                    ('X', 'M'),
                    ('M', 'G'),
                    ('G', 'R'),
                    ('R', 'X'),
                    ('M', 'R'),
                    ]

        # face centered cubic
        elif bravais_info['index'] == 2:
            special_points = {'G': [0., 0., 0.],
                              'K': [3. / 8., 3. / 8., 0.75],
                              'L': [0.5, 0.5, 0.5],
                              'U': [5. / 8., 0.25, 5. / 8.],
                              'W': [0.5, 0.25, 0.75],
                              'X': [0.5, 0., 0.5],
                              }
            path = [('G', 'X'),
                    ('X', 'W'),
                    ('W', 'K'),
                    ('K', 'G'),
                    ('G', 'L'),
                    ('L', 'U'),
                    ('U', 'W'),
                    ('W', 'L'),
                    ('L', 'K'),
                    ('U', 'X'),
                    ]

        # body centered cubic
        elif bravais_info['index'] == 3:
            special_points = {'G': [0., 0., 0.],
                              'H': [0.5, -0.5, 0.5],
                              'P': [0.25, 0.25, 0.25],
                              'N': [0., 0., 0.5],
                              }
            path = [('G', 'H'),
                    ('H', 'N'),
                    ('N', 'G'),
                    ('G', 'P'),
                    ('P', 'H'),
                    ('P', 'N'),
                    ]

        # Tetragonal
        elif bravais_info['index'] == 4:
            special_points = {'G': [0., 0., 0.],
                              'A': [0.5, 0.5, 0.5],
                              'M': [0.5, 0.5, 0.],
                              'R': [0., 0.5, 0.5],
                              'X': [0., 0.5, 0.],
                              'Z': [0., 0., 0.5],
                              }
            path = [('G', 'X'),
                    ('X', 'M'),
                    ('M', 'G'),
                    ('G', 'Z'),
                    ('Z', 'R'),
                    ('R', 'A'),
                    ('A', 'Z'),
                    ('X', 'R'),
                    ('M', 'A'),
                    ]

        # body centered tetragonal
        elif bravais_info['index'] == 5:
            if bravais_info['variation'] == 'bct1':
                # Body centered tetragonal bct1
                eta = bravais_info['extra']['eta']
                special_points = {'G': [0., 0., 0.],
                                  'M': [-0.5, 0.5, 0.5],
                                  'N': [0., 0.5, 0.],
                                  'P': [0.25, 0.25, 0.25],
                                  'X': [0., 0., 0.5],
                                  'Z': [eta, eta, -eta],
                                  'Z1': [-eta, 1. - eta, eta],
                                  }
                path = [('G', 'X'),
                        ('X', 'M'),
                        ('M', 'G'),
                        ('G', 'Z'),
                        ('Z', 'P'),
                        ('P', 'N'),
                        ('N', 'Z1'),
                        ('Z1', 'M'),
                        ('X', 'P'),
                        ]

            else:  # bct2
                # Body centered tetragonal bct2
                eta = bravais_info['extra']['eta']
                csi = bravais_info['extra']['csi']
                special_points = {
                    'G': [0., 0., 0.],
                    'N': [0., 0.5, 0.],
                    'P': [0.25, 0.25, 0.25],
                    'S': [-eta, eta, eta],
                    'S1': [eta, 1 - eta, -eta],
                    'X': [0., 0., 0.5],
                    'Y': [-csi, csi, 0.5],
                    'Y1': [0.5, 0.5, -csi],
                    'Z': [0.5, 0.5, -0.5],
                }
                path = [('G', 'X'),
                        ('X', 'Y'),
                        ('Y', 'S'),
                        ('S', 'G'),
                        ('G', 'Z'),
                        ('Z', 'S1'),
                        ('S1', 'N'),
                        ('N', 'P'),
                        ('P', 'Y1'),
                        ('Y1', 'Z'),
                        ('X', 'P'),
                        ]

        # orthorhombic
        elif bravais_info['index'] == 6:
            special_points = {'G': [0., 0., 0.],
                              'R': [0.5, 0.5, 0.5],
                              'S': [0.5, 0.5, 0.],
                              'T': [0., 0.5, 0.5],
                              'U': [0.5, 0., 0.5],
                              'X': [0.5, 0., 0.],
                              'Y': [0., 0.5, 0.],
                              'Z': [0., 0., 0.5],
                              }
            path = [('G', 'X'),
                    ('X', 'S'),
                    ('S', 'Y'),
                    ('Y', 'G'),
                    ('G', 'Z'),
                    ('Z', 'U'),
                    ('U', 'R'),
                    ('R', 'T'),
                    ('T', 'Z'),
                    ('Y', 'T'),
                    ('U', 'X'),
                    ('S', 'R'),
                    ]

        # face centered orthorhombic
        elif bravais_info['index'] == 7:
            if bravais_info['variation'] == 'orcf1':
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                special_points = {'G': [0., 0., 0.],
                                  'A': [0.5, 0.5 + csi, csi],
                                  'A1': [0.5, 0.5 - csi, 1. - csi],
                                  'L': [0.5, 0.5, 0.5],
                                  'T': [1., 0.5, 0.5],
                                  'X': [0., eta, eta],
                                  'X1': [1., 1. - eta, 1. - eta],
                                  'Y': [0.5, 0., 0.5],
                                  'Z': [0.5, 0.5, 0.],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'T'),
                        ('T', 'Z'),
                        ('Z', 'G'),
                        ('G', 'X'),
                        ('X', 'A1'),
                        ('A1', 'Y'),
                        ('T', 'X1'),
                        ('X', 'A'),
                        ('A', 'Z'),
                        ('L', 'G'),
                        ]

            elif bravais_info['variation'] == 'orcf2':
                eta = bravais_info['extra']['eta']
                dlt = bravais_info['extra']['dlt']
                phi = bravais_info['extra']['phi']
                special_points = {'G': [0., 0., 0.],
                                  'C': [0.5, 0.5 - eta, 1. - eta],
                                  'C1': [0.5, 0.5 + eta, eta],
                                  'D': [0.5 - dlt, 0.5, 1. - dlt],
                                  'D1': [0.5 + dlt, 0.5, dlt],
                                  'L': [0.5, 0.5, 0.5],
                                  'H': [1. - phi, 0.5 - phi, 0.5],
                                  'H1': [phi, 0.5 + phi, 0.5],
                                  'X': [0., 0.5, 0.5],
                                  'Y': [0.5, 0., 0.5],
                                  'Z': [0.5, 0.5, 0.],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'C'),
                        ('C', 'D'),
                        ('D', 'X'),
                        ('X', 'G'),
                        ('G', 'Z'),
                        ('Z', 'D1'),
                        ('D1', 'H'),
                        ('H', 'C'),
                        ('C1', 'Z'),
                        ('X', 'H1'),
                        ('H', 'Y'),
                        ('L', 'G'),
                        ]

            else:
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                special_points = {'G': [0., 0., 0.],
                                  'A': [0.5, 0.5 + csi, csi],
                                  'A1': [0.5, 0.5 - csi, 1. - csi],
                                  'L': [0.5, 0.5, 0.5],
                                  'T': [1., 0.5, 0.5],
                                  'X': [0., eta, eta],
                                  'X1': [1., 1. - eta, 1. - eta],
                                  'Y': [0.5, 0., 0.5],
                                  'Z': [0.5, 0.5, 0.],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'T'),
                        ('T', 'Z'),
                        ('Z', 'G'),
                        ('G', 'X'),
                        ('X', 'A1'),
                        ('A1', 'Y'),
                        ('X', 'A'),
                        ('A', 'Z'),
                        ('L', 'G'),
                        ]

        # Body centered orthorhombic
        elif bravais_info['index'] == 8:
            csi = bravais_info['extra']['csi']
            dlt = bravais_info['extra']['dlt']
            eta = bravais_info['extra']['eta']
            mu = bravais_info['extra']['mu']
            special_points = {'G': [0., 0., 0.],
                              'L': [-mu, mu, 0.5 - dlt],
                              'L1': [mu, -mu, 0.5 + dlt],
                              'L2': [0.5 - dlt, 0.5 + dlt, -mu],
                              'R': [0., 0.5, 0.],
                              'S': [0.5, 0., 0.],
                              'T': [0., 0., 0.5],
                              'W': [0.25, 0.25, 0.25],
                              'X': [-csi, csi, csi],
                              'X1': [csi, 1. - csi, -csi],
                              'Y': [eta, -eta, eta],
                              'Y1': [1. - eta, eta, -eta],
                              'Z': [0.5, 0.5, -0.5],
                              }
            path = [('G', 'X'),
                    ('X', 'L'),
                    ('L', 'T'),
                    ('T', 'W'),
                    ('W', 'R'),
                    ('R', 'X1'),
                    ('X1', 'Z'),
                    ('Z', 'G'),
                    ('G', 'Y'),
                    ('Y', 'S'),
                    ('S', 'W'),
                    ('L1', 'Y'),
                    ('Y1', 'Z'),
                    ]

        # C-centered orthorhombic
        elif bravais_info['index'] == 9:
            csi = bravais_info['extra']['csi']
            special_points = {'G': [0., 0., 0.],
                              'A': [csi, csi, 0.5],
                              'A1': [-csi, 1. - csi, 0.5],
                              'R': [0., 0.5, 0.5],
                              'S': [0., 0.5, 0.],
                              'T': [-0.5, 0.5, 0.5],
                              'X': [csi, csi, 0.],
                              'X1': [-csi, 1. - csi, 0.],
                              'Y': [-0.5, 0.5, 0.],
                              'Z': [0., 0., 0.5],
                              }
            path = [('G', 'X'),
                    ('X', 'S'),
                    ('S', 'R'),
                    ('R', 'A'),
                    ('A', 'Z'),
                    ('Z', 'G'),
                    ('G', 'Y'),
                    ('Y', 'X1'),
                    ('X1', 'A1'),
                    ('A1', 'T'),
                    ('T', 'Y'),
                    ('Z', 'T'),
                    ]

        # Hexagonal
        elif bravais_info['index'] == 10:
            special_points = {'G': [0., 0., 0.],
                              'A': [0., 0., 0.5],
                              'H': [1. / 3., 1. / 3., 0.5],
                              'K': [1. / 3., 1. / 3., 0.],
                              'L': [0.5, 0., 0.5],
                              'M': [0.5, 0., 0.],
                              }
            path = [('G', 'M'),
                    ('M', 'K'),
                    ('K', 'G'),
                    ('G', 'A'),
                    ('A', 'L'),
                    ('L', 'H'),
                    ('H', 'A'),
                    ('L', 'M'),
                    ('K', 'H'),
                    ]

        # rhombohedral
        elif bravais_info['index'] == 11:
            if bravais_info['variation'] == 'rhl1':
                eta = bravais_info['extra']['eta']
                nu = bravais_info['extra']['nu']
                special_points = {'G': [0., 0., 0.],
                                  'B': [eta, 0.5, 1. - eta],
                                  'B1': [0.5, 1. - eta, eta - 1.],
                                  'F': [0.5, 0.5, 0.],
                                  'L': [0.5, 0., 0.],
                                  'L1': [0., 0., -0.5],
                                  'P': [eta, nu, nu],
                                  'P1': [1. - nu, 1. - nu, 1. - eta],
                                  'P2': [nu, nu, eta - 1.],
                                  'Q': [1. - nu, nu, 0.],
                                  'X': [nu, 0., -nu],
                                  'Z': [0.5, 0.5, 0.5],
                                  }
                path = [('G', 'L'),
                        ('L', 'B1'),
                        ('B', 'Z'),
                        ('Z', 'G'),
                        ('G', 'X'),
                        ('Q', 'F'),
                        ('F', 'P1'),
                        ('P1', 'Z'),
                        ('L', 'P'),
                        ]

            else:  # Rhombohedral rhl2
                eta = bravais_info['extra']['eta']
                nu = bravais_info['extra']['nu']
                special_points = {'G': [0., 0., 0.],
                                  'F': [0.5, -0.5, 0.],
                                  'L': [0.5, 0., 0.],
                                  'P': [1. - nu, -nu, 1. - nu],
                                  'P1': [nu, nu - 1., nu - 1.],
                                  'Q': [eta, eta, eta],
                                  'Q1': [1. - eta, -eta, -eta],
                                  'Z': [0.5, -0.5, 0.5],
                                  }
                path = [('G', 'P'),
                        ('P', 'Z'),
                        ('Z', 'Q'),
                        ('Q', 'G'),
                        ('G', 'F'),
                        ('F', 'P1'),
                        ('P1', 'Q1'),
                        ('Q1', 'L'),
                        ('L', 'Z'),
                        ]

        # monoclinic
        elif bravais_info['index'] == 12:
            eta = bravais_info['extra']['eta']
            nu = bravais_info['extra']['nu']
            special_points = {'G': [0., 0., 0.],
                              'A': [0.5, 0.5, 0.],
                              'C': [0., 0.5, 0.5],
                              'D': [0.5, 0., 0.5],
                              'D1': [0.5, 0., -0.5],
                              'E': [0.5, 0.5, 0.5],
                              'H': [0., eta, 1. - nu],
                              'H1': [0., 1. - eta, nu],
                              'H2': [0., eta, -nu],
                              'M': [0.5, eta, 1. - nu],
                              'M1': [0.5, 1. - eta, nu],
                              'M2': [0.5, eta, -nu],
                              'X': [0., 0.5, 0.],
                              'Y': [0., 0., 0.5],
                              'Y1': [0., 0., -0.5],
                              'Z': [0.5, 0., 0.],
                              }
            path = [('G', 'Y'),
                    ('Y', 'H'),
                    ('H', 'C'),
                    ('C', 'E'),
                    ('E', 'M1'),
                    ('M1', 'A'),
                    ('A', 'X'),
                    ('X', 'H1'),
                    ('M', 'D'),
                    ('D', 'Z'),
                    ('Y', 'D'),
                    ]

        elif bravais_info['index'] == 13:
            if bravais_info['variation'] == 'mclc1':
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                psi = bravais_info['extra']['psi']
                phi = bravais_info['extra']['phi']
                special_points = {'G': [0., 0., 0.],
                                  'N': [0.5, 0., 0.],
                                  'N1': [0., -0.5, 0.],
                                  'F': [1. - csi, 1. - csi, 1. - eta],
                                  'F1': [csi, csi, eta],
                                  'F2': [csi, -csi, 1. - eta],
                                  'F3': [1. - csi, -csi, 1. - eta],
                                  'I': [phi, 1. - phi, 0.5],
                                  'I1': [1. - phi, phi - 1., 0.5],
                                  'L': [0.5, 0.5, 0.5],
                                  'M': [0.5, 0., 0.5],
                                  'X': [1. - psi, psi - 1., 0.],
                                  'X1': [psi, 1. - psi, 0.],
                                  'X2': [psi - 1., -psi, 0.],
                                  'Y': [0.5, 0.5, 0.],
                                  'Y1': [-0.5, -0.5, 0.],
                                  'Z': [0., 0., 0.5],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'F'),
                        ('F', 'L'),
                        ('L', 'I'),
                        ('I1', 'Z'),
                        ('Z', 'F1'),
                        ('Y', 'X1'),
                        ('X', 'G'),
                        ('G', 'N'),
                        ('M', 'G'),
                        ]

            elif bravais_info['variation'] == 'mclc2':
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                psi = bravais_info['extra']['psi']
                phi = bravais_info['extra']['phi']
                special_points = {'G': [0., 0., 0.],
                                  'N': [0.5, 0., 0.],
                                  'N1': [0., -0.5, 0.],
                                  'F': [1. - csi, 1. - csi, 1. - eta],
                                  'F1': [csi, csi, eta],
                                  'F2': [csi, -csi, 1. - eta],
                                  'F3': [1. - csi, -csi, 1. - eta],
                                  'I': [phi, 1. - phi, 0.5],
                                  'I1': [1. - phi, phi - 1., 0.5],
                                  'L': [0.5, 0.5, 0.5],
                                  'M': [0.5, 0., 0.5],
                                  'X': [1. - psi, psi - 1., 0.],
                                  'X1': [psi, 1. - psi, 0.],
                                  'X2': [psi - 1., -psi, 0.],
                                  'Y': [0.5, 0.5, 0.],
                                  'Y1': [-0.5, -0.5, 0.],
                                  'Z': [0., 0., 0.5],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'F'),
                        ('F', 'L'),
                        ('L', 'I'),
                        ('I1', 'Z'),
                        ('Z', 'F1'),
                        ('N', 'G'),
                        ('G', 'M'),
                        ]

            elif bravais_info['variation'] == 'mclc3':
                mu = bravais_info['extra']['mu']
                dlt = bravais_info['extra']['dlt']
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                phi = bravais_info['extra']['phi']
                psi = bravais_info['extra']['psi']
                special_points = {
                    'G': [0., 0., 0.],
                    'F': [1. - phi, 1 - phi, 1. - psi],
                    'F1': [phi, phi - 1., psi],
                    'F2': [1. - phi, -phi, 1. - psi],
                    'H': [csi, csi, eta],
                    'H1': [1. - csi, -csi, 1. - eta],
                    'H2': [-csi, -csi, 1. - eta],
                    'I': [0.5, -0.5, 0.5],
                    'M': [0.5, 0., 0.5],
                    'N': [0.5, 0., 0.],
                    'N1': [0., -0.5, 0.],
                    'X': [0.5, -0.5, 0.],
                    'Y': [mu, mu, dlt],
                    'Y1': [1. - mu, -mu, -dlt],
                    'Y2': [-mu, -mu, -dlt],
                    'Y3': [mu, mu - 1., dlt],
                    'Z': [0., 0., 0.5],
                }
                path = [('G', 'Y'),
                        ('Y', 'F'),
                        ('F', 'H'),
                        ('H', 'Z'),
                        ('Z', 'I'),
                        ('I', 'F1'),
                        ('H1', 'Y1'),
                        ('Y1', 'X'),
                        ('X', 'F'),
                        ('G', 'N'),
                        ('M', 'G'),
                        ]

            elif bravais_info['variation'] == 'mclc4':
                mu = bravais_info['extra']['mu']
                dlt = bravais_info['extra']['dlt']
                csi = bravais_info['extra']['csi']
                eta = bravais_info['extra']['eta']
                phi = bravais_info['extra']['phi']
                psi = bravais_info['extra']['psi']
                special_points = {'G': [0., 0., 0.],
                                  'F': [1. - phi, 1 - phi, 1. - psi],
                                  'F1': [phi, phi - 1., psi],
                                  'F2': [1. - phi, -phi, 1. - psi],
                                  'H': [csi, csi, eta],
                                  'H1': [1. - csi, -csi, 1. - eta],
                                  'H2': [-csi, -csi, 1. - eta],
                                  'I': [0.5, -0.5, 0.5],
                                  'M': [0.5, 0., 0.5],
                                  'N': [0.5, 0., 0.],
                                  'N1': [0., -0.5, 0.],
                                  'X': [0.5, -0.5, 0.],
                                  'Y': [mu, mu, dlt],
                                  'Y1': [1. - mu, -mu, -dlt],
                                  'Y2': [-mu, -mu, -dlt],
                                  'Y3': [mu, mu - 1., dlt],
                                  'Z': [0., 0., 0.5],
                                  }
                path = [('G', 'Y'),
                        ('Y', 'F'),
                        ('F', 'H'),
                        ('H', 'Z'),
                        ('Z', 'I'),
                        ('H1', 'Y1'),
                        ('Y1', 'X'),
                        ('X', 'G'),
                        ('G', 'N'),
                        ('M', 'G'),
                        ]

            else:
                csi = bravais_info['extra']['csi']
                mu = bravais_info['extra']['mu']
                omg = bravais_info['extra']['omg']
                eta = bravais_info['extra']['eta']
                nu = bravais_info['extra']['nu']
                dlt = bravais_info['extra']['dlt']
                rho = bravais_info['extra']['rho']
                special_points = {
                    'G': [0., 0., 0.],
                    'F': [nu, nu, omg],
                    'F1': [1. - nu, 1. - nu, 1. - omg],
                    'F2': [nu, nu - 1., omg],
                    'H': [csi, csi, eta],
                    'H1': [1. - csi, -csi, 1. - eta],
                    'H2': [-csi, -csi, 1. - eta],
                    'I': [rho, 1. - rho, 0.5],
                    'I1': [1. - rho, rho - 1., 0.5],
                    'L': [0.5, 0.5, 0.5],
                    'M': [0.5, 0., 0.5],
                    'N': [0.5, 0., 0.],
                    'N1': [0., -0.5, 0.],
                    'X': [0.5, -0.5, 0.],
                    'Y': [mu, mu, dlt],
                    'Y1': [1. - mu, -mu, -dlt],
                    'Y2': [-mu, -mu, -dlt],
                    'Y3': [mu, mu - 1., dlt],
                    'Z': [0., 0., 0.5],
                }
                path = [('G', 'Y'),
                        ('Y', 'F'),
                        ('F', 'L'),
                        ('L', 'I'),
                        ('I1', 'Z'),
                        ('Z', 'H'),
                        ('H', 'F1'),
                        ('H1', 'Y1'),
                        ('Y1', 'X'),
                        ('X', 'G'),
                        ('G', 'N'),
                        ('M', 'G'),
                        ]

        # triclinic
        elif bravais_info['index'] == 14:
            if bravais_info['variation'] == 'tri1a' or bravais_info['variation'] == 'tri2a':
                special_points = {'G': [0.0, 0.0, 0.0],
                                  'L': [0.5, 0.5, 0.0],
                                  'M': [0.0, 0.5, 0.5],
                                  'N': [0.5, 0.0, 0.5],
                                  'R': [0.5, 0.5, 0.5],
                                  'X': [0.5, 0.0, 0.0],
                                  'Y': [0.0, 0.5, 0.0],
                                  'Z': [0.0, 0.0, 0.5],
                                  }
                path = [('X', 'G'),
                        ('G', 'Y'),
                        ('L', 'G'),
                        ('G', 'Z'),
                        ('N', 'G'),
                        ('G', 'M'),
                        ('R', 'G'),
                        ]

            else:
                special_points = {'G': [0.0, 0.0, 0.0],
                                  'L': [0.5, -0.5, 0.0],
                                  'M': [0.0, 0.0, 0.5],
                                  'N': [-0.5, -0.5, 0.5],
                                  'R': [0.0, -0.5, 0.5],
                                  'X': [0.0, -0.5, 0.0],
                                  'Y': [0.5, 0.0, 0.0],
                                  'Z': [-0.5, 0.0, 0.5],
                                  }
                path = [('X', 'G'),
                        ('G', 'Y'),
                        ('L', 'G'),
                        ('G', 'Z'),
                        ('N', 'G'),
                        ('G', 'M'),
                        ('R', 'G'),
                        ]

    elif dimension == 2:
        # 2D case: 5 Bravais lattices
        if bravais_info['index'] == 1:
            # square
            special_points = {'G': [0., 0., 0.],
                              'M': [0.5, 0.5, 0.],
                              'X': [0.5, 0., 0.],
                              }
            path = [('G', 'X'),
                    ('X', 'M'),
                    ('M', 'G'),
                    ]

        elif bravais_info['index'] == 2:
            # (primitive) rectangular
            special_points = {'G': [0., 0., 0.],
                              'X': [0.5, 0., 0.],
                              'Y': [0., 0.5, 0.],
                              'S': [0.5, 0.5, 0.],
                              }
            path = [('G', 'X'),
                    ('X', 'S'),
                    ('S', 'Y'),
                    ('Y', 'G'),
                    ]

        elif bravais_info['index'] == 3:
            # centered rectangular (rhombic)
            # TODO: this looks quite different from the in-plane part of the
            # 3D C-centered orthorhombic lattice, which is strange...

            # NOTE: special points below are in (b1, b2) fractional
            # coordinates (primitive reciprocal cell) as for the rest.
            # Ramirez & Bohn gave them initially in (s1=b1+b2, s2=-b1+b2)
            # coordinates, i.e. using the conventional reciprocal cell.
            special_points = {'G': [0., 0., 0.],
                              'X': [0.5, 0.5, 0.],
                              'Y1': [0.25, 0.75, 0.],
                              'Y': [-0.25, 0.25, 0.],  # typo in p. 404 of Ramirez & Bohm (should be Y=(0,1/4))
                              'C': [0., 0.5, 0.],
                              }
            path = [('Y1', 'X'),
                    ('X', 'G'),
                    ('G', 'Y'),
                    ('Y', 'C'),
                    ]

        elif bravais_info['index'] == 4:
            # hexagonal
            special_points = {'G': [0., 0., 0.],
                              'M': [0.5, 0., 0.],
                              'K': [1. / 3., 1. / 3., 0.],
                              }
            path = [('G', 'M'),
                    ('M', 'K'),
                    ('K', 'G'),
                    ]

        elif bravais_info['index'] == 5:
            # oblique
            # NOTE: only end-points are high-symmetry points (not the path
            # in-between)
            special_points = {'G': [0., 0., 0.],
                              'X': [0.5, 0., 0.],
                              'Y': [0., 0.5, 0.],
                              'A': [0.5, 0.5, 0.],
                              }
            path = [('X', 'G'),
                    ('G', 'Y'),
                    ('A', 'G'),
                    ]

    elif dimension == 1:
        # 1D case: 1 Bravais lattice
        special_points = {'G': [0., 0., 0.],
                          'X': [0.5, 0., 0.],
                          }
        path = [('G', 'X'),
                ]

    elif dimension == 0:
        # 0D case: 1 Bravais lattice, only Gamma point, no path
        special_points = {'G': [0., 0., 0.],
                          }
        path = [('G', 'G'),
                ]

    permutation = bravais_info['permutation']

    def permute(x, permutation):
        # return new_x such that new_x[i]=x[permutation[i]]
        return [x[int(p)] for p in permutation]

    def invpermute(permutation):
        # return the inverse of permutation
        return [permutation.index(i) for i in range(3)]

    the_special_points = {}
    for k in special_points.iterkeys():
        # NOTE: this originally returned the inverse of the permutation, but was later changed to permutation
        the_special_points[k] = permute(special_points[k], permutation)

    # output crystal or cartesian
    if cartesian:
        the_abs_special_points = {}
        for k in the_special_points.iterkeys():
            the_abs_special_points[k] = change_reference(
                reciprocal_cell, numpy.array(the_special_points[k]), to_cartesian=True
            )

        return the_abs_special_points, path, bravais_info
    else:
        return the_special_points, path, bravais_info


#!/usr/bin/env python
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import os
import itertools
try:
    from pathlib import Path
except ImportError:
    from pathlib2 import Path

def find_path(root, dir_name):
    path = Path(os.path.abspath(root))
    for p in itertools.chain([path], path.parents):
        directory = p / dir_name
        if directory.is_dir():
            return directory
    raise OSError('No directory found')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.backends.general.abstractqueries import AbstractQueryManager


class QueryManagerSQLA(AbstractQueryManager):
    """
    SQLAlchemy implementation of custom queries, for efficiency reasons
    """

    def get_creation_statistics(
            self,
            user_pk=None
    ):
        """
        Return a dictionary with the statistics of node creation, summarized by day,
        optimized for the Django backend.

        :note: Days when no nodes were created are not present in the returned `ctime_by_day` dictionary.

        :param user_pk: If None (default), return statistics for all users.
            If user pk is specified, return only the statistics for the given user.

        :return: a dictionary as
            follows::

                {
                   "total": TOTAL_NUM_OF_NODES,
                   "types": {TYPESTRING1: count, TYPESTRING2: count, ...},
                   "ctime_by_day": {'YYYY-MMM-DD': count, ...}

            where in `ctime_by_day` the key is a string in the format 'YYYY-MM-DD' and the value is
            an integer with the number of nodes created that day.
        """
        import sqlalchemy as sa
        import aiida.backends.sqlalchemy
        from aiida.backends.sqlalchemy import models as m

        # Get the session (uses internally aldjemy - so, sqlalchemy) also for the Djsite backend
        s = aiida.backends.sqlalchemy.get_scoped_session()

        retdict = {}

        total_query = s.query(m.node.DbNode)
        types_query = s.query(m.node.DbNode.type.label('typestring'),
                sa.func.count(m.node.DbNode.id))
        stat_query = s.query(sa.func.date_trunc('day', m.node.DbNode.ctime).label('cday'),
                           sa.func.count(m.node.DbNode.id))

        if user_pk is not None:
            total_query = total_query.filter(m.node.DbNode.user_id == user_pk)
            types_query = types_query.filter(m.node.DbNode.user_id == user_pk)
            stat_query = stat_query.filter(m.node.DbNode.user_id == user_pk)

        # Total number of nodes
        retdict["total"] = total_query.count()

        # Nodes per type
        retdict["types"] = dict(types_query.group_by('typestring').all())

        # Nodes created per day
        stat = stat_query.group_by('cday').order_by('cday').all()

        ctime_by_day = {_[0].strftime('%Y-%m-%d'): _[1] for _ in stat}
        retdict["ctime_by_day"] = ctime_by_day

        return retdict
        # Still not containing all dates





# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import aiida.common
from aiida.common.exceptions import InternalError
from aiida.common.extendeddicts import FixedFieldsAttributeDict

from aiida.transport.transport import Transport
from aiida.transport.util import TransportFactory


def copy_from_remote_to_remote(transportsource, transportdestination, remotesource, remotedestination, **kwargs):
    """
    Copy files or folders from a remote computer to another remote computer.

    :param transportsource: transport to be used for the source computer
    :param transportdestination: transport to be used for the destination computer
    :param str remotesource: path to the remote source directory / file
    :param str remotedestination: path to the remote destination directory / file
    :param kwargs: keyword parameters passed to the final put,
        except for 'dereference' that is passed to the initial get

    .. note:: it uses the method transportsource.copy_from_remote_to_remote
    """
    transportsource.copy_from_remote_to_remote(transportdestination, remotesource, remotedestination, **kwargs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Tests for the NWChem input plugins.
"""
from aiida.backends.testbase import AiidaTestCase
from aiida.transport import TransportFactory
from aiida.common.exceptions import NotExistent




class TestComputer(AiidaTestCase):

    def test_get_transport(self):
        """
        Test the get_transport method of Computer
        """
        import tempfile
        from aiida.orm import Computer
        from aiida.orm.backend import construct_backend
        backend = construct_backend()

        new_comp = Computer(name='bbb',
                                hostname='localhost',
                                transport_type='local',
                                scheduler_type='direct',
                                workdir='/tmp/aiida')
        new_comp.store()

        # Configure the computer - no parameters for local transport
        authinfo = backend.authinfos.create(computer=new_comp, user=backend.users.get_automatic_user())
        authinfo.store()

        transport = new_comp.get_transport()

        # It's on localhost, so I see files that I create
        with transport:
            with tempfile.NamedTemporaryFile() as f:
                self.assertEquals(transport.isfile(f.name), True)
            # Here the file should have been deleted
            self.assertEquals(transport.isfile(f.name), False)

    def test_delete(self):
        from aiida.orm import Computer
        new_comp = Computer(name='aaa',
                                hostname='aaa',
                                transport_type='local',
                                scheduler_type='pbspro',
                                workdir='/tmp/aiida')
        new_comp.store()

        comp_pk = new_comp.pk

        check_computer = Computer.get(comp_pk)
        self.assertEquals(comp_pk, check_computer.pk)
        
        from aiida.orm.computer import delete_computer
        delete_computer(pk=comp_pk)

        with self.assertRaises(NotExistent):
            Computer.get(comp_pk)




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Test ssh plugin on localhost
"""
import unittest
import logging

import aiida.transport
import aiida.transport.transport
import paramiko
from aiida.transport.plugins.ssh import SshTransport

# This will be used by test_all_plugins

plugin_transport = SshTransport(machine='localhost', timeout=30, load_system_host_keys=True, key_policy='AutoAddPolicy')


class TestBasicConnection(unittest.TestCase):
    """
    Test basic connections.
    """

    def test_closed_connection_ssh(self):
        with self.assertRaises(aiida.transport.transport.TransportInternalError):
            t = SshTransport(machine='localhost')
            t._exec_command_internal('ls')

    def test_closed_connection_sftp(self):
        with self.assertRaises(aiida.transport.transport.TransportInternalError):
            t = SshTransport(machine='localhost')
            t.listdir()

    def test_invalid_param(self):
        with self.assertRaises(ValueError):
            SshTransport(machine='localhost', invalid_param=True)

    def test_auto_add_policy(self):
        with SshTransport(machine='localhost', timeout=30, load_system_host_keys=True, key_policy='AutoAddPolicy'):
            pass

    def test_no_host_key(self):
        # Disable logging to avoid output during test
        logging.disable(logging.ERROR)

        with self.assertRaises(paramiko.SSHException):
            with SshTransport(machine='localhost', timeout=30, load_system_host_keys=False):
                pass

        # Reset logging level
        logging.disable(logging.NOTSET)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################


def delete_nodes(pks, follow_calls=False, follow_returns=False, 
                 dry_run=False, force=False, disable_checks=False, verbosity=0):
    """
    Delete nodes by a list of pks

    :note: The script will also delete all children calculations generated from the specified nodes.

    :param pks: a list of the PKs of the nodes to delete
    :param bool follow_calls: Follow calls
    :param bool follow_returns:
        Follow returns. This is a very dangerous option, since anything returned by a workflow might have
        been used as input in many other calculations. Use with care, and never combine with force.
    :param bool dry_run: Do not delete, a dry run, with statistics printed according to verbosity levels.
    :param bool force: Do not ask for confirmation to delete nodes.
    :param bool disable_checks:
        If True, will not check whether calculations are losing created data or called instances.
        If checks are disabled, also logging is disabled.
    :param bool force: Do not ask for confirmation to delete nodes.
    :param int verbosity:
        The verbosity levels, 0 prints nothing, 1 prints just sums and total, 2 prints individual nodes.
    """

    from aiida.orm.querybuilder import QueryBuilder
    from aiida.common.links import LinkType
    from aiida.orm.node import Node
    from aiida.orm.calculation import Calculation
    from aiida.orm.data import Data
    from aiida.orm import load_node
    from aiida.orm.backend import construct_backend
    from aiida.backends.utils import delete_nodes_and_connections

    backend = construct_backend()
    user_email = backend.users.get_automatic_user().email

    if not pks:
        # If I was passed an empty list, I don't to anything
        # I prefer checking explicitly, an empty set might be problematic for the queries done below.
        if verbosity:
            print "Nothing to delete"
        return

    # The following code is just for the querying of downwards provenance.
    # Ideally, there should be a module to interface with, but this is the solution for now.
    # By only dealing with ids, and keeping track of what has been already
    # visited in the query, there's good performance and no infinite loops.
    link_types_to_follow = [LinkType.CREATE.value, LinkType.INPUT.value]
    if follow_calls:
        link_types_to_follow.append(LinkType.CALL.value)
    if follow_returns:
        link_types_to_follow.append(LinkType.RETURN.value)

    edge_filters = {'type': {'in': link_types_to_follow}}

    # Operational set always includes the recently (in the last iteration added) nodes.
    operational_set = set().union(set(pks))  # Union to copy the set!
    pks_set_to_delete = set().union(set(pks))
    while operational_set:
        # new_pks_set are the the pks of all nodes that are connected to the operational node set
        # with the links specified.
        new_pks_set = set([i for i, in QueryBuilder().append(
            Node, filters={'id': {'in': operational_set}}).append(
            Node, project='id', edge_filters=edge_filters).iterall()])
        # The operational set is only those pks that haven't been yet put into the pks_set_to_delete.
        operational_set = new_pks_set.difference(pks_set_to_delete)

        # I add these pks in the pks_set_to_delete with a union
        pks_set_to_delete = pks_set_to_delete.union(new_pks_set)

    if verbosity > 0:
        print "I {} delete {} node{}".format(
            'would' if dry_run else 'will',
            len(pks_set_to_delete),
            's' if len(pks_set_to_delete) > 1 else '')
        if verbosity > 1:
            qb = QueryBuilder().append(Node, filters={'id': {'in': pks_set_to_delete}},
                                       project=('uuid', 'id', 'type', 'label'))
            print "The nodes I {} delete:".format('would' if dry_run else 'will')
            for uuid, pk, type_string, label in qb.iterall():
                try:
                    short_type_string = type_string.split('.')[-2]
                except IndexError:
                    short_type_string = type_string
                print "   {} {} {} {}".format(uuid, pk, short_type_string, label)


    # Here I am checking whether I am deleting
    ## A data instance without also deleting the creator, which brakes relationship between a calculation and its data
    ## A calculation instance that was called, without also deleting the caller.

    if not disable_checks:
        called_qb = QueryBuilder()
        called_qb.append(Calculation, filters={'id': {'!in': pks_set_to_delete}}, project='id')
        called_qb.append(Calculation, project='type', edge_project='label',
                         filters={'id': {'in': pks_set_to_delete}},
                         edge_filters={'type': {'==': LinkType.CALL.value}})
        caller_to_called2delete = called_qb.all()

        if verbosity > 0 and caller_to_called2delete:
            calculation_pks_losing_called = set(zip(*caller_to_called2delete)[0])
            print "\n{} calculation{} {} lose at least one called instance".format(
                    len(calculation_pks_losing_called),
                    's' if len(calculation_pks_losing_called) > 1 else '',
                    'would' if dry_run else 'will')
            if verbosity > 1:
                print "These are the calculations that {} lose a called instance:".format('would' if dry_run else 'will')
                for calc_losing_called_pk in calculation_pks_losing_called:
                    print '  ', load_node(calc_losing_called_pk)

        created_qb = QueryBuilder()
        created_qb.append(Calculation, filters={'id': {'!in': pks_set_to_delete}}, project='id')
        created_qb.append(Data, project='type', edge_project='label',
                          filters={'id': {'in': pks_set_to_delete}},
                          edge_filters={'type': {'==': LinkType.CREATE.value}})

        creator_to_created2delete = created_qb.all()
        if verbosity > 0 and creator_to_created2delete:
            calculation_pks_losing_created = set(zip(*creator_to_created2delete)[0])
            print "\n{} calculation{} {} lose at least one created data-instance".format(
                len(calculation_pks_losing_created),
                's' if len(calculation_pks_losing_created) > 1 else '',
                'would' if dry_run else 'will')
            if verbosity > 1:
                print "These are the calculations that {} lose a created data-instance:".format(
                    'would' if dry_run else 'will')
                for calc_losing_created_pk in calculation_pks_losing_created:
                    print '  ', load_node(calc_losing_created_pk)

    if dry_run:
        if verbosity > 0:
            print "\nThis was a dry run, exiting without deleting anything"
        return

    # Asking for user confirmation here
    if force:
        pass
    else:
        print "YOU ARE ABOUT TO DELETE {} NODES! THIS CANNOT BE UNDONE!".format(len(pks_set_to_delete))
        if raw_input("Shall I continue? [Y/N] ").lower() != 'y':
            print "Exiting without deleting"
            return

    # Recover the list of folders to delete before actually deleting
    # the nodes.  I will delete the folders only later, so that if
    # there is a problem during the deletion of the nodes in
    # the DB, I don't delete the folders
    folders = [load_node(_).folder for _ in pks_set_to_delete]

    delete_nodes_and_connections(pks_set_to_delete)

    if not disable_checks:
        # I pass now to the log the information for calculations losing created data or called instances
        for calc_pk, calc_type_string, link_label in caller_to_called2delete:
            calc = load_node(calc_pk)
            calc.logger.warning("User {} deleted "
                                "an instance of type {} "
                                "called with the label {} "
                                "by this calculation".format(
                user_email, calc_type_string, link_label))

        for calc_pk, data_type_string, link_label in creator_to_created2delete:
            calc = load_node(calc_pk)
            calc.logger.warning("User {} deleted "
                                "an instance of type {} "
                                "created with the label {} "
                                "by this calculation".format(
                user_email, data_type_string, link_label))

    # If we are here, we managed to delete the entries from the DB.
    # I can now delete the folders
    for f in folders:
        f.erase()




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import numpy
from aiida.orm.data import BaseType
from aiida.orm.data import to_aiida_type


class Bool(BaseType):
    """
    Class to store booleans as AiiDA nodes
    """
    _type = bool

    def __int__(self):
        return int(bool(self))

    # Python 2
    def __nonzero__(self):
        return self.__bool__()

    # Python 3
    def __bool__(self):
        return self.value


@to_aiida_type.register(bool)
@to_aiida_type.register(numpy.bool_)
def _(value):
    return Bool(value)


def get_true_node():
    """
    Return a Bool Data node, with value True

    Cannot be done as a singleton in the module, because it would be generated
    at import time, with the risk that (e.g. in the tests, or at the very first use
    of AiiDA) a user is not yet defined in the DB (but a user is mandatory in the
    DB before you can create new Nodes in AiiDA).
    """
    TRUE = Bool(typevalue=(bool, True))
    return TRUE


def get_false_node():
    """
    Return a Bool Data node, with value False

    Cannot be done as a singleton in the module, because it would be generated
    at import time, with the risk that (e.g. in the tests, or at the very first use
    of AiiDA) a user is not yet defined in the DB (but a user is mandatory in the
    DB before you can create new Nodes in AiiDA).
    """
    FALSE = Bool(typevalue=(bool, False))
    return FALSE



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""create account table

Revision ID: 470e57bc0936
Revises: 
Create Date: 2017-06-16 17:53:10.920345

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '470e57bc0936'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'account',
        sa.Column('id', sa.Integer, primary_key=True),
        sa.Column('name', sa.String(50), nullable=False),
        sa.Column('description', sa.Unicode(200)),
    )

def downgrade():
    op.drop_table('account')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.backends.testbase import AiidaTestCase

import aiida.orm
from aiida.orm.data.int import Int
from aiida.orm.data.str import Str
from aiida.work.launch import run, run_get_node
from aiida.work.test_utils import DummyProcess


class TestRun(AiidaTestCase):
    def setUp(self):
        super(TestRun, self).setUp()

    def tearDown(self):
        super(TestRun, self).tearDown()

    def test_run(self):
        inputs = {'a': Int(2), 'b': Str('test')}
        result = run(DummyProcess, **inputs)

    def test_run_get_node(self):
        inputs = {'a': Int(2), 'b': Str('test')}
        result, node = run_get_node(DummyProcess, **inputs)
        self.assertIsInstance(node, aiida.orm.Calculation)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import aiida
from aiida.work import Process
from aiida.backends.testbase import AiidaTestCase
from aiida.orm.calculation.job.simpleplugins.templatereplacer import TemplatereplacerCalculation


class TestJobProcess(AiidaTestCase):
    def setUp(self):
        super(TestJobProcess, self).setUp()
        self.assertIsNone(Process.current())

    def tearDown(self):
        super(TestJobProcess, self).tearDown()
        self.assertIsNone(Process.current())

    def test_class_loader(self):
        templatereplacer_process = aiida.work.JobProcess.build(TemplatereplacerCalculation)
        loader = aiida.work.get_object_loader()

        class_name = loader.identify_object(templatereplacer_process)
        loaded_class = loader.load_object(class_name)

        self.assertEqual(templatereplacer_process.__name__, loaded_class.__name__)
        self.assertEqual(class_name, loader.identify_object(loaded_class))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import os
import django
from aiida.common.log import get_dblogger_extra


def load_dbenv(profile=None):
    """
    Load the database environment (Django) and perform some checks.

    :param profile: the string with the profile to use. If not specified,
        use the default one specified in the AiiDA configuration file.
    """
    _load_dbenv_noschemacheck(profile)
    # Check schema version and the existence of the needed tables
    check_schema_version()


def _load_dbenv_noschemacheck(profile):
    """
    Load the database environment (Django) WITHOUT CHECKING THE SCHEMA VERSION.

    :param profile: the string with the profile to use. If not specified,
        use the default one specified in the AiiDA configuration file.

    This should ONLY be used internally, inside load_dbenv, and for schema
    migrations. DO NOT USE OTHERWISE!
    """
    # This function does not use process and profile because they are read
    # from global variables (set before by load_profile) inside the
    # djsite.settings.settings module.
    os.environ['DJANGO_SETTINGS_MODULE'] = 'aiida.backends.djsite.settings.settings'
    django.setup()


def get_log_messages(obj):
    from aiida.backends.djsite.db.models import DbLog
    import json

    extra = get_dblogger_extra(obj)
    # convert to list, too
    log_messages = list(DbLog.objects.filter(**extra).order_by('time').values(
        'loggername', 'levelname', 'message', 'metadata', 'time'))

    # deserialize metadata
    for log in log_messages:
        log.update({'metadata': json.loads(log['metadata'])})

    return log_messages


_aiida_autouser_cache = None


def long_field_length():
    """
    Return the length of "long" fields.
    This is used, for instance, for the 'key' field of attributes.
    This returns 1024 typically, but it returns 255 if the backend is mysql.

    :note: Call this function only AFTER having called load_dbenv!
    """
    # One should not load directly settings because there are checks inside
    # for the current profile. However, this function is going to be called
    # only after having loaded load_dbenv, so there should be no problem
    from django.conf import settings

    if 'mysql' in settings.DATABASES['default']['ENGINE']:
        return 255
    else:
        return 1024


def check_schema_version():
    """
    Check if the version stored in the database is the same of the version
    of the code.

    :note: if the DbSetting table does not exist, this function does not
      fail. The reason is to avoid to have problems before running the first
      migrate call.

    :note: if no version is found, the version is set to the version of the
      code. This is useful to have the code automatically set the DB version
      at the first code execution.

    :raise ConfigurationError: if the two schema versions do not match.
      Otherwise, just return.
    """
    import os
    import aiida.backends.djsite.db.models
    from aiida.backends.utils import get_current_profile
    from django.db import connection
    from aiida.common.exceptions import ConfigurationError

    # Do not do anything if the table does not exist yet
    if 'db_dbsetting' not in connection.introspection.table_names():
        return

    code_schema_version = aiida.backends.djsite.db.models.SCHEMA_VERSION
    db_schema_version = get_db_schema_version()

    if db_schema_version is None:
        # No code schema defined yet, I set it to the code version
        set_db_schema_version(code_schema_version)
        db_schema_version = get_db_schema_version()

    filepath_utils = os.path.abspath(__file__)
    filepath_manage = os.path.join(os.path.dirname(filepath_utils), 'manage.py')

    if code_schema_version != db_schema_version:
        raise ConfigurationError(
            "The code schema version is {}, but the version stored in the "
            "database (DbSetting table) is {}, stopping.\n"
            "To migrate the database to the current version, run the following commands:"
            "\n  verdi daemon stop\n  python {} --aiida-profile={} migrate".
                format(
                code_schema_version,
                db_schema_version,
                filepath_manage,
                get_current_profile()
            )
        )


def set_db_schema_version(version):
    """
    Set the schema version stored in the DB. Use only if you know what
    you are doing.
    """
    from aiida.backends.utils import set_global_setting
    return set_global_setting(
        'db|schemaversion', version,
        description="The version of the schema used in this database.")


def get_db_schema_version():
    """
    Get the current schema version stored in the DB. Return None if
    it is not stored.
    """
    from aiida.backends.utils import get_global_setting
    try:
        return get_global_setting('db|schemaversion')
    except KeyError:
        return None


def delete_nodes_and_connections_django(pks_to_delete):
    """
    Delete all nodes corresponding to pks in the input.
    :param pks_to_delete: A list, tuple or set of pks that should be deleted.
    """
    from django.db import transaction
    from django.db.models import Q
    from aiida.backends.djsite.db import models
    with transaction.atomic():
        # Delete all links pointing to or from a given node
        models.DbLink.objects.filter(
            Q(input__in=pks_to_delete) |
            Q(output__in=pks_to_delete)).delete()
        # now delete nodes
        models.DbNode.objects.filter(pk__in=pks_to_delete).delete()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Tests for specific subclasses of Data
"""

from aiida.backends.testbase import AiidaTestCase



### Here comparisons are defined #####################################
### Each comparison has to be a function with name
### _comparison_COMPARISONNAME
### and accepting three parameters:
### - testclass: the testing class, with the proper AssertXXX methods
### - the dbdata value, i.e. the value parsed by the test
### - comparisondata, the values specified by the user for comparison;
###   they typically contain a 'value', and possibly other keys for
###   more advanced keys
def _comparison_AlmostEqual(testclass, dbdata, comparisondata):
    """
    Compare two numbers (or a list of numbers) to check that
    they are all almost equal (within a default precision of 7 digits)
    """
    value = comparisondata['value']
    if isinstance(dbdata, (list, tuple)) and isinstance(value, (list, tuple)):
        testclass.assertEqual(len(dbdata), len(value))
        for i in range(0, len(dbdata)):
            testclass.assertAlmostEqual(dbdata[i], value[i])
    else:
        testclass.assertAlmostEqual(dbdata, value)


def _comparison_Equal(testclass, dbdata, comparisondata):
    """
    Compare two objects to see if they are equal
    """
    testclass.assertEqual(dbdata, comparisondata['value'])


def _comparison_LengthEqual(testclass, dbdata, comparisondata):
    """
    Check if the length of the object is equal to the value specified
    """
    testclass.assertEqual(len(dbdata), comparisondata['value'])


### End of comparison definition #####################################

def output_test(pk, testname, skip_uuids_from_inputs=[]):
    """
    This is the function that should be used to create a new test from an
    existing calculation.

    It is possible to simplify the file removing unwanted nodes.

    :param pk: PK of Calculation, used for test
    :param testname: the name of this test, used to create a new folder.
        The folder name will be of the form test_PLUGIN_TESTNAME,
        with PLUGIN substituted by the plugin name, with dots replaced by
        underscores. Testname can contain only digits, letters and underscores.
    :param skip_uuids_from_inputs: a list of UUIDs of input nodes to be
        skipped
    """
    import os
    import json

    from aiida.common.folders import Folder
    from aiida.orm import JobCalculation
    from aiida.orm.utils import load_node
    from aiida.orm.importexport import export_tree
    c = load_node(pk, parent_class=JobCalculation)
    outfolder = "test_{}_{}".format(
        c.get_parser_name().replace('.', '_'),
        testname)

    if not is_valid_folder_name(outfolder):
        raise ValueError("The testname is invalid; it can contain only "
                         "letters, digits or underscores")

    if os.path.exists(outfolder):
        raise ValueError("Out folder '{}' already exists".format(outfolder))

    inputs = []
    for node in c.get_inputs():
        if node.uuid not in skip_uuids_from_inputs:
            inputs.append(node.dbnode)

    folder = Folder(outfolder)
    to_export = [c.dbnode] + inputs
    try:
        to_export.append(c.out.retrieved.dbnode)
    except AttributeError:
        raise ValueError("No output retrieved node; without it, we cannot "
                         "test the parser!")
    export_tree(to_export, folder=folder,
                also_parents=False,
                also_calc_outputs=False)

    # Create an empty checks file
    with open(os.path.join(outfolder, '_aiida_checks.json'), 'w') as f:
        json.dump({}, f)

    for path, dirlist, filelist in os.walk(outfolder):
        if len(dirlist) == 0 and len(filelist) == 0:
            with open("{}/.gitignore".format(path), 'w') as f:
                f.write("# This is a placeholder file, used to make git "
                        "store an empty folder")
                f.flush()


def is_valid_folder_name(name):
    """
    Return True if the string (that will be the folder name of each subtest)
    is a valid name for a test function: it should start with ``test_``, and
    contain only letters, digits or underscores.
    """
    import string

    if not name.startswith('test_'):
        return False

    # Remove valid characters, see if anything remains
    bad_characters = name.translate(None, string.letters + string.digits + '_')
    if bad_characters:
        return False

    return True

class TestParsers(AiidaTestCase):
    """
    This class dynamically finds all tests in a given subfolder, and loads
    them as different tests.
    """
    # To have both the "default" error message from assertXXX, and the
    # msg specified by us
    longMessage = True

    def read_test(self, outfolder):
        import os
        import importlib
        import json

        from aiida.orm import JobCalculation
        from aiida.orm.utils import load_node
        from aiida.orm.importexport import import_data

        imported = import_data(outfolder,
                               ignore_unknown_nodes=True, silent=True)

        calc = None
        for _, pk in imported['aiida.backends.djsite.db.models.DbNode']['new']:
            c = load_node(pk)
            if issubclass(c.__class__, JobCalculation):
                calc = c
                break

        retrieved = calc.out.retrieved

        try:
            with open(os.path.join(outfolder, '_aiida_checks.json')) as f:
                tests = json.load(f)
        except IOError:
            raise ValueError("This test does not provide a check file!")
        except ValueError:
            raise ValueError("This test does provide a check file, but it cannot "
                             "be JSON-decoded!")

        mod_path = 'aiida.backends.tests.parser_tests.{}'.format(
            os.path.split(outfolder)[1])

        skip_test = False
        try:
            m = importlib.import_module(mod_path)
            skip_test = m.skip_condition()
        except Exception:
            pass

        if skip_test:
            raise SkipTestException

        return calc, {'retrieved': retrieved}, tests

    @classmethod
    def return_base_test(cls, folder):
        from inspect import isfunction

        def base_test(self):
            try:
                calc, retrieved_nodes, tests = self.read_test(folder)
            except SkipTestException:
                return None
            Parser = calc.get_parserclass()
            if Parser is None:
                raise NotImplementedError
            else:
                parser = Parser(calc)
                successful, new_nodes_tuple = parser.parse_with_retrieved(
                    retrieved_nodes)
                self.assertTrue(successful, msg="The parser did not succeed")
                parsed_output_nodes = dict(new_nodes_tuple)

                # All main keys: name of nodes that should be present
                for test_node_name in tests:
                    try:
                        test_node = parsed_output_nodes[test_node_name]
                    except KeyError:
                        raise AssertionError("Output node '{}' expected but "
                                             "not found".format(test_node_name))

                    # Each subkey: attribute to check
                    # attr_test is the name of the attribute
                    for attr_test in tests[test_node_name]:
                        try:
                            dbdata = test_node.get_attr(attr_test)
                        except AttributeError:
                            raise AssertionError("Attribute '{}' not found in "
                                                 "parsed node '{}'".format(
                                attr_test,
                                test_node_name))
                        # Test data from the JSON
                        attr_test_listtests = tests[test_node_name][attr_test]
                        for test_number, attr_test_data in enumerate(
                                attr_test_listtests, start=1):
                            try:
                                comparison = attr_test_data.pop('comparison')
                            except KeyError as e:
                                raise ValueError(
                                    "Missing '{}' in the '{}' field "
                                    "in '{}' in "
                                    "the test file".format(e.message,
                                                           attr_test,
                                                           test_node_name))

                            try:
                                comparison_test = globals()[
                                    "_comparison_{}".format(comparison)]
                            except KeyError:
                                raise ValueError(
                                    "Unsupported '{}' comparison in "
                                    "the '{}' field in '{}' in "
                                    "the test file".format(comparison,
                                                           attr_test,
                                                           test_node_name))

                            if not isfunction(comparison_test):
                                raise TypeError(
                                    "Internal error: the variable _comparison_{} is not a "
                                    "function!".format(comparison))

                            try:
                                comparison_test(testclass=self, dbdata=dbdata,
                                                comparisondata=attr_test_data)
                            except Exception as e:
                                # I change both the message and the 'args'
                                # (apparently, args[0] is used by str(e))
                                # Probably, a 'better' way should be found to do this!
                                e.message = "Failed test #{} for {}->{}: {}".format(
                                    test_number, test_node_name, attr_test,
                                    e.message)
                                if e.args:
                                    e.args = tuple(
                                        ["Failed test #{} for {}->{}: {}".format(
                                            test_number,
                                            test_node_name,
                                            attr_test,
                                            e.args[0])]
                                        + list(e.args[1:]))
                                raise e

        return base_test

    class __metaclass__(type):
        """
        Some python black magic to dynamically create tests
        """

        def __new__(cls, name, bases, attrs):
            import os

            newcls = type.__new__(cls, name, bases, attrs)

            file_folder = os.path.split(__file__)[0]
            parser_test_folder = os.path.join(file_folder, 'parser_tests')
            if os.path.isdir(parser_test_folder):
                for f in os.listdir(parser_test_folder):
                    absf = os.path.abspath(os.path.join(parser_test_folder, f))
                    if is_valid_folder_name(f) and os.path.isdir(absf):
                        function_name = f
                        setattr(newcls, function_name,
                                newcls.return_base_test(absf))

            return newcls


class SkipTestException(Exception):
    pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import os

from aiida.backends import settings
from aiida.common import setup


CONFIG_DIR = setup.AIIDA_CONFIG_FOLDER
DAEMON_DIR = 'daemon'
DAEMON_LOG_DIR = 'daemon/log'

CIRCUS_PID_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.pid')
DAEMON_PID_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'aiida-{}.pid')
CIRCUS_LOG_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_LOG_DIR, 'circus-{}.log')
DAEMON_LOG_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_LOG_DIR, 'aiida-{}.log')
CIRCUS_PORT_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.port')
CIRCUS_SOCKET_FILE_TEMPATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.sockets')
CIRCUS_CONTROLLER_SOCKET_TEMPLATE = 'circus.c.sock'
CIRCUS_PUBSUB_SOCKET_TEMPLATE = 'circus.p.sock'
CIRCUS_STATS_SOCKET_TEMPLATE = 'circus.s.sock'


def get_current_profile_name():
    """
    Return the currently configured profile name or if not set, the default profile
    """
    return settings.AIIDADB_PROFILE or setup.get_default_profile()


def get_current_profile_config():
    """
    Return the configuration of the currently active profile or if not set, the default profile
    """
    return setup.get_profile_config(get_current_profile_name())


class ProfileConfig(object):
    """
    Convenience class that loads the current profile name and corresponding configuration and
    can subsequently provide profile specific information
    """

    _RMQ_PREFIX = 'aiida-{uuid}'

    def __init__(self, profile_name=None):
        """
        Construct the ProfileConfig for the given profile_name or retrieve it from
        the backend settings
        """
        if not profile_name:
            self.profile_name = get_current_profile_name()
            self.profile_config = get_current_profile_config()
        else:
            self.profile_name = profile_name
            self.profile_config = setup.get_profile_config(profile_name)

    @property
    def profile_uuid(self):
        return self.profile_config[setup.PROFILE_UUID_KEY]

    @property
    def rmq_prefix(self):
        return self._RMQ_PREFIX.format(uuid=self.profile_uuid)

    @property
    def filepaths(self):
        return {
            'circus': {
                'log': CIRCUS_LOG_FILE_TEMPLATE.format(self.profile_name),
                'pid': CIRCUS_PID_FILE_TEMPLATE.format(self.profile_name),
                'port': CIRCUS_PORT_FILE_TEMPLATE.format(self.profile_name),
                'socket': {
                    'file': CIRCUS_SOCKET_FILE_TEMPATE.format(self.profile_name),
                    'controller': CIRCUS_CONTROLLER_SOCKET_TEMPLATE,
                    'pubsub': CIRCUS_PUBSUB_SOCKET_TEMPLATE,
                    'stats': CIRCUS_STATS_SOCKET_TEMPLATE,
                }
            },
            'daemon': {
                'log': DAEMON_LOG_FILE_TEMPLATE.format(self.profile_name),
                'pid': DAEMON_PID_FILE_TEMPLATE.format(self.profile_name),
            }
        }



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import unittest



class FoldersTest(unittest.TestCase):
    """
    Tests for the Folder class.
    """

    def test_unicode(self):
        """
        Check that there are no exceptions raised when 
        using unicode folders.
        """
        from aiida.common.folders import Folder
        import os, tempfile

        tmpsource = tempfile.mkdtemp()
        tmpdest = tempfile.mkdtemp()
        with open(os.path.join(tmpsource, "sąžininga"), 'w') as f:
            f.write("test")
        with open(os.path.join(tmpsource, "žąsis"), 'w') as f:
            f.write("test")
        fd = Folder(tmpdest)
        fd.insert_path(tmpsource, "destination")
        fd.insert_path(tmpsource, u"šaltinis")

        fd = Folder(os.path.join(tmpsource, u"šaltinis"))
        fd.insert_path(tmpsource, "destination")
        fd.insert_path(tmpdest, u"kitas-šaltinis")


    def test_get_abs_path_without_limit(self):
        from aiida.common.folders import Folder

        fd = Folder('/tmp')
        # Should not raise any exception
        self.assertEquals(fd.get_abs_path('test_file.txt'),
                          '/tmp/test_file.txt')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.backends.testbase import AiidaTestCase


class TestQueryBuilderSQLA(AiidaTestCase):
    def test_clsf_sqla(self):
        from aiida.orm import Group, User, Computer, Node, Data, Calculation
        from aiida.backends.sqlalchemy.models.node import DbNode
        from aiida.backends.sqlalchemy.models.group import DbGroup
        from aiida.backends.sqlalchemy.models.user import DbUser
        from aiida.backends.sqlalchemy.models.computer import DbComputer

        from aiida.orm.querybuilder import QueryBuilder

        qb = QueryBuilder()
        for AiidaCls, ORMCls, typestr in zip(
                (Group, User, Computer, Node, Data, Calculation),
                (DbGroup, DbUser, DbComputer, DbNode, DbNode, DbNode),
                (None, None, None, Node._query_type_string, Data._query_type_string, Calculation._query_type_string)):
            cls, clstype, query_type_string = qb._get_ormclass(AiidaCls, None)

            self.assertEqual(cls, ORMCls)
            self.assertEqual(query_type_string, typestr)


class QueryBuilderLimitOffsetsTestSQLA(AiidaTestCase):

    def test_ordering_limits_offsets_of_results_for_SQLA(self):
        from aiida.orm import Node
        from aiida.orm.querybuilder import QueryBuilder
        # Creating 10 nodes with an attribute that can be ordered
        for i in range(10):
            n = Node()
            n._set_attr('foo', i)
            n.store()
        qb = QueryBuilder().append(
            Node, project='attributes.foo'
        ).order_by(
            {Node: {'attributes.foo': {'cast': 'i'}}}
        )
        res = list(zip(*qb.all())[0])
        self.assertEqual(res, range(10))

        # Now applying an offset:
        qb.offset(5)
        res = list(zip(*qb.all())[0])
        self.assertEqual(res, range(5, 10))

        # Now also applying a limit:
        qb.limit(3)
        res = list(zip(*qb.all())[0])
        self.assertEqual(res, range(5, 8))
